# Copyright (c) 2015 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

'''
Cinder Volume driver for Fujitsu ETERNUS DX S3 series.
'''


import time
import hashlib
import base64
from oslo_config import cfg
from xml.etree.ElementTree import parse
from cinder import context
from cinder import exception
from cinder import utils
from oslo_utils import units
from cinder.volume import volume_types
from cinder.volume import qos_specs
from cinder.volume.configuration import Configuration
from cinder.volume.drivers.fujitsu import eternus_dx_scheduler
from oslo_log import log as logging
from cinder.volume.drivers.fujitsu import eternus_dx_lockutils
from oslo_service import loopingcall
from oslo_concurrency import processutils
import six
from cinder.i18n import _LE, _LI, _LW

LOG = logging.getLogger(__name__)

try:
    import pywbem
except:
    msg = (_LE('import pywbem failed!!'
               'pywbem is necessary for this volume driver.'))
    LOG.error(msg)
    raise exception.VolumeBackendAPIException(data=msg)

VOL_PREFIX = "FJosv_"
RAIDGROUP = 2
TPPOOL = 5
SNAPOPC = 4
OPC = 5
RETURN_TO_RESOURCEPOOL = 19
DETACH = 8
INITIALIZED = 2
UNSYNCHRONIZED = 3
BROKEN = 5
PREPARED = 11
REPL = "FUJITSU_ReplicationService"
STOR_CONF = "FUJITSU_StorageConfigurationService"
CTRL_CONF = "FUJITSU_ControllerConfigurationService"
STOR_HWID = "FUJITSU_StorageHardwareIDManagementService"

UNDEF_MSG = 'Undefined Error!!'
JOB_RETRIES = 60
JOB_INTERVAL_SEC = 10
# error code keyword
VOLUMENAME_IN_USE = 32788L
COPYSESSION_NOT_EXIST = 32793L
LUNAME_IN_USE = 4102L
LUNAME_NOT_EXIST = 4097L  # only for InvokeMethod(HidePaths)
EC_REC = 3
FJ_ETERNUS_DX_OPT_list = [cfg.StrOpt('cinder_eternus_config_file',
                                     default='/etc/cinder/cinder_fujitsu_'
                                             'eternus_dx.xml',
                                     help='config file for cinder fujitsu'
                                          '_eternus_dx volume driver')]
POOL_TYPE_dic = {RAIDGROUP: 'RAID_GROUP',
                 TPPOOL: 'Thinporvisioning_POOL'
                 }

OPERATION_dic = {SNAPOPC: RETURN_TO_RESOURCEPOOL,
                 OPC: DETACH,
                 EC_REC: DETACH}

RETCODE_dic = {'0': 'Success',
               '1': 'Method Not Supported',
               '4': 'Failed',
               '5': 'Invalid Parameter',
               '4096': 'Method Parameters Checked - Job Started',
               '4097': 'Size Not Supported',
               '4101': 'Target/initiator combination already exposed',
               '4102': 'Requested logical unit number in use',
               '32769': 'Maximum number of Logical Volume in'
                        ' a RAID group has been reached',
               '32770': 'Maximum number of Logical Volume in'
                        ' the storage device has been reached',
               '32771': 'Maximum number of registered Host WWN'
                        ' has been reached',
               '32772': 'Maximum number of affinity group has been reached',
               '32773': 'Maximum number of host affinity has been reached',
               '32785': 'The RAID group is in busy state',
               '32786': 'The Logical Volume is in busy state',
               '32787': 'The device is in busy state',
               '32788': 'Element Name is in use',
               '32792': 'No Copy License',
               '32793': 'Session is not exist',
               '32796': 'Quick Format Error',
               '32801': 'The CA port is in invalid setting',
               '32802': 'The Logical Volume is Mainframe volume',
               '32803': 'The RAID group is not operative',
               '32804': 'The Logical Volume is not operative',
               '32808': 'No Thin Provisioning License',
               '32809': 'The Logical Element is ODX volume',
               '32811': 'This operation cannot be performed to the NAS'
                        ' resources',
               '32812': 'This operation cannot be performed to the Storage'
                        ' Cluster resources',
               '32816': 'Fatal error generic',
               '35302': 'Invalid LogicalElement',
               '35304': 'LogicalElement state error',
               '35316': 'Multi-hop error',
               '35318': 'Maximum number of multi-hop has been reached',
               '35324': 'RAID is broken',
               '35331': 'Maximum number of session has been reached(per'
                        ' device)',
               '35333': 'Maximum number of session has been reached(per'
                        ' SourceElement)',
               '35334': 'Maximum number of session has been reached(per'
                        ' TargetElement)',
               '35335': 'Maximum number of Snapshot generation has been'
                        ' reached (per SourceElement)',
               '35346': 'Copy table size is not setup',
               '35347': 'Copy table size is not enough'}


class FJDXCommon(object):
    '''
    Common code that does not depend on protocol.
    '''

    backend_ip = ""
    VERSION = "1.3.0"
    stats = {'driver_version': VERSION,
             'free_capacity_gb': 0,
             'reserved_percentage': 0,
             'storage_protocol': None,
             'total_capacity_gb': 0,
             'vendor_name': 'FUJITSU',
             'QoS_support': False,
             'volume_backend_name': None}

    def __init__(self, prtcl, configuration=None):
        '''
        Constructor
        '''
        self.protocol = prtcl
        self.configuration = configuration
        self.configuration.append_config_values(FJ_ETERNUS_DX_OPT_list)

        if prtcl == 'iSCSI':
            # get iSCSI ipaddress from driver configuration file
            self.configuration.iscsi_ip_address = \
                self._get_drvcfg('EternusISCSIIP')
        # end of if

        self.conn = self._get_eternus_connection()
        backend_ip = self._get_drvcfg('EternusIP')
        self.__setattr__('backend_ip', backend_ip)
        self.scheduler = eternus_dx_scheduler.FJDXScheduler(self.configuration)

        return

    def create_volume(self, volume):
        '''
        create volume on ETERNUS
        '''
        LOG.debug('create_volume,Enter method')

        # conversion of the unit. GB -> MB -> KB -> B
        volumesize = int(volume['size']) * units.Gi

        # create to volumename on ETERNUS from cinder VolumeID
        volumename = self._create_volume_name(volume['id'])
        LOG.debug('create_volume,volumename:%(volumename)s,'
                  'volumesize:%(volumesize)u'
                  % {'volumename': volumename,
                     'volumesize': volumesize})

        self.conn = self._get_eternus_connection()

        # get suitable pool form config pools
        # Existence check the pool
        pool_instance = self._get_pool_from_scheduler(
            volume['size'], use_tpp=True)
        if pool_instance is None:
            msg = (_LE('create_volume,'
                       'not found suitable pool form config pools.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        # set pooltype
        pool_path = pool_instance['path']
        if 'RSP' in pool_path['InstanceID']:
            pooltype = RAIDGROUP
        else:
            pooltype = TPPOOL
        # end of if

        configservice = self._find_eternus_service(STOR_CONF)
        if configservice is None:
            msg = (_LE('create_volume,volume:%(volume)s,'
                       'volumename:%(volumename)s,'
                       'eternus_pool:%(eternus_pool)s,'
                       'Error!! Storage Configuration Service is None.')
                   % {'volume': volume,
                      'volumename': volumename,
                      'eternus_pool': pool_instance['name']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('create_volume,CreateOrModifyElementFromStoragePool,'
                  'ConfigService:%(service)s,'
                  'ElementName:%(volumename)s,'
                  'InPool:%(eternus_pool)s,'
                  'ElementType:%(pooltype)u,'
                  'Size:%(volumesize)u'
                  % {'service': configservice,
                     'volumename': volumename,
                     'eternus_pool': pool_instance['name'],
                     'pooltype': pooltype,
                     'volumesize': volumesize})

        # Invoke method for create volume
        rc, errordesc, job = self._exec_eternus_service(
            'CreateOrModifyElementFromStoragePool',
            configservice,
            ElementName=volumename,
            InPool=pool_path,
            ElementType=pywbem.Uint16(pooltype),
            Size=pywbem.Uint64(volumesize))

        if rc == VOLUMENAME_IN_USE:  # Element Name is in use
            msg = (_LW('create_volume,'
                       'volumename:%(volumename)s,'
                       'Return code:%(rc)lu,'
                       'Error:%(errordesc)s')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.warn(msg)
            vol_instance = self._find_lun(volume)
            element = vol_instance
        elif rc != 0L:
            msg = (_LE('create_volume,'
                       'volumename:%(volumename)s,'
                       'Return code:%(rc)lu,'
                       'Error:%(errordesc)s')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        else:
            element = job['TheElement']
        # end of if

        # get eternus model for metadata
        # systemname = systemnamelist[0]['IdentifyingNumber']
        # ex) ET092DC4511133A10
        try:
            systemnamelist = \
                self._enum_eternus_instances('FUJITSU_StorageProduct')
        except:
            msg = (_LE('create_volume,'
                       'volume:%(volume)s,'
                       'EnumerateInstances,'
                       'cannot connect to ETERNUS.')
                   % {'volume': volume})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('create_volume,'
                  'volumename:%(volumename)s,'
                  'Return code:%(rc)lu,'
                  'Error:%(errordesc)s,'
                  'Backend:%(backend)s,'
                  'Pool Name:%(eternus_pool)s,'
                  'Pool Type:%(pooltype)s,'
                  'Exit method'
                  % {'volumename': volumename,
                     'rc': rc,
                     'errordesc': errordesc,
                     'backend': systemnamelist[0]['IdentifyingNumber'],
                     'eternus_pool': pool_instance['name'],
                     'pooltype': POOL_TYPE_dic[pooltype]})

        element_path = {}
        # create return value
        element_path['classname'] = element.classname
        element_path['keybindings'] = \
            {'CreationClassName': element['CreationClassName'],
             'SystemName': element['SystemName'],
             'DeviceID': element['DeviceID'],
             'SystemCreationClassName': element['SystemCreationClassName']}

        volume_no = "0x" + element['DeviceID'][24:28]

        metadata = {'FJ_Backend': systemnamelist[0]['IdentifyingNumber'],
                    'FJ_Volume_Name': volumename,
                    'FJ_Volume_No': volume_no,
                    'FJ_Pool_Name': pool_instance['name'],
                    'FJ_Pool_Type': POOL_TYPE_dic[pooltype]}

        return (element_path, metadata)

    def create_volume_from_snapshot(self, volume, snapshot):
        '''
        Creates a volume from a snapshot
        '''
        LOG.debug('create_volume_from_snapshot,Enter method')

        source_volume_instance = self._find_lun(snapshot)

        # Existence check the source volume
        if source_volume_instance is None:
            msg = (_LE('create_volume_from_snapshot,'
                       'Source Volume is not exist in ETERNUS.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        # create volume for destination of cloned volume
        (element_path, metadata) = self.create_volume(volume)
        target_volume_instancename = \
            self._create_eternus_instance_name(element_path['classname'],
                                               element_path['keybindings'])
        target_volume_instance = \
            self._get_eternus_instance(target_volume_instancename)
        self._create_local_cloned_volume(volume, target_volume_instance,
                                         snapshot, source_volume_instance)

        LOG.debug('create_volume_from_snapshot,Exit method')

        return (element_path, metadata)

    def create_cloned_volume(self, volume, src_vref):
        '''
        Create clone of the specified volume.
        '''
        LOG.debug('create_cloned_volume,Enter method')

        source_volume_instance = self._find_lun(src_vref)

        if source_volume_instance is None:
            msg = (_LE('create_cloned_volume,'
                       'Source Volume is not exist in ETERNUS.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        (element_path, metadata) = self.create_volume(volume)
        target_volume_instancename = \
            self._create_eternus_instance_name(
                element_path['classname'], element_path['keybindings'])
        target_volume_instance = \
            self._get_eternus_instance(target_volume_instancename)

        self._create_local_cloned_volume(volume, target_volume_instance,
                                         src_vref, source_volume_instance)

        LOG.debug('create_cloned_volume,Exit method')
        return (element_path, metadata)

    @eternus_dx_lockutils.FJDXLockutils('vol', 'cinder-', True)
    def _create_local_cloned_volume(self, volume, target_volume_instance,
                                    src_vref, source_volume_instance):
        '''
        Create local clone of the specified volume.
        '''
        LOG.debug('_create_local_cloned_volume,Enter method')

        self.conn = self._get_eternus_connection()
        s_volumename = self._create_volume_name(src_vref['id'])
        t_volumename = self._create_volume_name(volume['id'])

        LOG.debug('_create_local_cloned_volume,'
                  'volumename:%(volumename)s,'
                  'sourcevolumename:%(sourcevolumename)s,'
                  'source volume instance:%(source_volume_instance)s,'
                  % {'volumename': t_volumename,
                     'sourcevolumename': s_volumename,
                     'source_volume_instance':
                         str(source_volume_instance.path)})

        # get replicationservice for CreateElementReplica
        repservice = self._find_eternus_service(REPL)

        if repservice is None:
            msg = (_LE('_create_local_cloned_volume,'
                       'Replication Service not found'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        # Invoke method for create cloned volume from volume
        rc, errordesc, job = self._exec_eternus_service(
            'CreateElementReplica',
            repservice,
            SyncType=pywbem.Uint16(8),
            SourceElement=source_volume_instance.path,
            TargetElement=target_volume_instance.path)

        if rc != 0L:
            msg = (_LE('_create_local_cloned_volume,'
                       'volumename:%(volumename)s,'
                       'sourcevolumename:%(sourcevolumename)s,'
                       'source volume instance:%(source_volume_instance)s,'
                       'target volume instance:%(target_volume_instance)s,'
                       'Return code:%(rc)lu,'
                       'Error:%(errordesc)s')
                   % {'volumename': t_volumename,
                      'sourcevolumename': s_volumename,
                      'source_volume_instance':
                          str(source_volume_instance.path),
                      'target_volume_instance':
                          str(target_volume_instance.path),
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('_create_local_cloned_volume,Exit method')
        return

    def delete_volume(self, volume):
        '''
        Delete volume on ETERNUS.
        '''
        LOG.debug('delete_volume,Enter method')

        valid = self._delete_volume_setting(volume)

        if valid is False:
            return
        # end of if

        vol_instance = self._find_lun(volume)
        if vol_instance is None:
            return
        # end of if
        self._delete_volume(vol_instance)

        LOG.debug('delete_volume,Exit method')
        return

    @eternus_dx_lockutils.FJDXLockutils('vol', 'cinder-', True)
    def _delete_volume_setting(self, volume):
        '''
        Delete volume setting ( HostAffinity, CopySession) on ETERNUS.
        '''
        LOG.debug('_delete_volume_setting,Enter method')

        # Existence check the volume
        volumename = self._create_volume_name(volume['id'])
        vol_instance = self._find_lun(volume)

        if vol_instance is None:
            LOG.info(_LI('_delete_volume_setting,volumename:%(volumename)s,'
                         'volume not found on ETERNUS.'
                         'delete only management data on cinder database.')
                     % {'volumename': volumename})
            return False
        # end of if

        # delete host-affinity setting remained by unexpected error
        self._unmap_lun(volume, None, force=True)

        # check copy session relating to target volume
        cpsessionlist = self._find_copysession(vol_instance)
        delete_copysession_list = []
        wait_copysession_list = []

        for cpsession in cpsessionlist:
            LOG.debug('_delete_volume_setting,'
                      'volumename: %(volumename)s, '
                      'cpsession: %(cpsession)s'
                      % {'volumename': volumename,
                         'cpsession': cpsession})

            if cpsession['SyncedElement'] == vol_instance.path:
                # Copy destination : other_volume --(copy)--> vol_instance
                delete_copysession_list.append(cpsession)
            elif cpsession['SystemElement'] == vol_instance.path:
                # Copy source : vol_instance --(copy)--> other volume
                wait_copysession_list.append(cpsession)
            # end of if
        # end of for cpsessionlist

        for cpsession in wait_copysession_list:
            self._wait_for_copy_complete(cpsession)
        # end of for wait_copysession_list

        for cpsession in delete_copysession_list:
            self._delete_copysession(cpsession)
        # end of for delete_copysession_list

        LOG.debug('_delete_volume_setting,Exit method')
        return True

    @eternus_dx_lockutils.FJDXLockutils('vol', 'cinder-', True)
    def _delete_volume(self, vol_instance):
        '''
        Delete volume on ETERNUS.
        '''
        LOG.debug('_delete_volume,Enter method')

        self.conn = self._get_eternus_connection()
        volumename = vol_instance['ElementName']

        configservice = self._find_eternus_service(STOR_CONF)
        if configservice is None:
            msg = (_LE('delete_volume,volumename:%(volumename)s,'
                       'Storage Configuration Service not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('delete_volume,volumename:%(volumename)s,'
                  'vol_instance:%(vol_instance)s,'
                  'Method: ReturnToStoragePool'
                  % {'volumename': volumename,
                     'vol_instance': str(vol_instance.path)})

        # Invoke method for delete volume
        rc, errordesc, job = self._exec_eternus_service(
            'ReturnToStoragePool',
            configservice,
            TheElement=vol_instance.path)

        if rc != 0L:
            msg = (_LE('delete_volume,volumename:%(volumename)s,'
                       'Return code:%(rc)lu,'
                       'Error:%(errordesc)s')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('delete_volume,volumename:%(volumename)s,'
                  'Return code:%(rc)lu,'
                  'Error:%(errordesc)s,'
                  'Exit Method'
                  % {'volumename': volumename,
                     'rc': rc,
                     'errordesc': errordesc})
        return

    @eternus_dx_lockutils.FJDXLockutils('vol', 'cinder-', True)
    def create_snapshot(self, snapshot):
        '''
        create snapshot using SnapOPC
        '''
        LOG.debug('create_snapshot,Enter method')

        element = None
        element_path = {}

        snapshotname = snapshot['name']
        volumename = snapshot['volume_name']
        vol_id = snapshot['volume_id']
        volume = snapshot['volume']
        d_volumename = self._create_volume_name(snapshot['id'])
        s_volumename = self._create_volume_name(vol_id)
        vol_instance = self._find_lun(volume)
        repservice = self._find_eternus_service(REPL)
        self.conn = self._get_eternus_connection()

        # Existence check the volume
        if vol_instance is None:
            # source volume is not found on ETERNUS
            msg = (_LE('create_snapshot,'
                       'volumename on ETERNUS:%(s_volumename)s,'
                       'source volume is not found on ETERNUS.')
                   % {'s_volumename': s_volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        if repservice is None:
            msg = (_LE('create_snapshot,'
                       'volumename:%(volumename)s,'
                       'Replication Service not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        # get poolname from driver configuration file
        eternus_pool = self._get_snap_pool_name(snapshot)
        # Existence check the pool
        pool = self._find_pool(eternus_pool)
        if pool is None:
            msg = (_LE('create_snapshot,'
                       'eternus_pool:%(eternus_pool)s,'
                       'not found.')
                   % {'eternus_pool': eternus_pool})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('create_snapshot,'
                  'snapshotname:%(snapshotname)s,'
                  'source volume name:%(volumename)s,'
                  'vol_instance.path:%(vol_instance)s,'
                  'dest_volumename:%(d_volumename)s,'
                  'pool:%(pool)s,'
                  'Invoke CreateElementReplica'
                  % {'snapshotname': snapshotname,
                     'volumename': volumename,
                     'vol_instance': str(vol_instance.path),
                     'd_volumename': d_volumename,
                     'pool': pool})

        # Invoke method for create snapshot
        rc, errordesc, job = self._exec_eternus_service(
            'CreateElementReplica',
            repservice,
            ElementName=d_volumename,
            TargetPool=pool,
            SyncType=pywbem.Uint16(7),
            SourceElement=vol_instance.path)

        if rc != 0L:
            msg = (_LE('create_snapshot,'
                       'snapshotname:%(snapshotname)s,'
                       'source volume name:%(volumename)s,'
                       'vol_instance.path:%(vol_instance)s,'
                       'dest_volumename:%(d_volumename)s,'
                       'pool:%(pool)s,'
                       'Return code:%(rc)lu,'
                       'Error:%(errordesc)s')
                   % {'snapshotname': snapshotname,
                      'volumename': volumename,
                      'vol_instance': str(vol_instance.path),
                      'd_volumename': d_volumename,
                      'pool': pool,
                      'rc': rc,
                      'errordesc': errordesc})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        else:
            element = job['TargetElement']
        # end of if

        LOG.debug('create_snapshot,volumename:%(volumename)s,'
                  'Return code:%(rc)lu,'
                  'Error:%(errordesc)s,'
                  'Exit Method'
                  % {'volumename': volumename,
                     'rc': rc,
                     'errordesc': errordesc})

        # create return value
        if element is not None:
            element_path['classname'] = element.classname
            element_path['keybindings'] = \
                {'CreationClassName': element['CreationClassName'],
                 'SystemName': element['SystemName'],
                 'DeviceID': element['DeviceID'],
                 'SystemCreationClassName':
                     element['SystemCreationClassName']}
        # end of if

        return element_path

    def delete_snapshot(self, snapshot):
        '''
        delete snapshot
        '''
        LOG.debug('delete_snapshot,Enter method')

        self.delete_volume(snapshot)

        LOG.debug('delete_snapshot,Exit method')
        return

    def initialize_connection(self, volume, connector):
        '''
        Allow connection to connector and return connection info.
        '''
        LOG.debug('initialize_connection,Enter method')

        vol_instance = self._find_lun(volume)
        # Existence check the volume
        if vol_instance is None:
            # volume does not found
            msg = (_LE('initialize_connection,'
                       'volume:%(volume)s,'
                       'Volume not found.')
                   % {'volume': volume['name']})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        self.conn = self._get_eternus_connection()
        target_portlist = self._get_target_port()
        mapdata = self._get_mapdata(vol_instance, connector, target_portlist)

        if mapdata is not None:
            # volume is already mapped
            target_lun = mapdata['target_lun'] if 'target_lun' in mapdata \
                else mapdata['target_luns'][0]
            msg = (_LI('initialize_connection,'
                       'volume:%(volume)s,'
                       'target_lun:%(target_lun)s,'
                       'Volume is already mapped.')
                   % {'volume': volume['name'],
                      'target_lun': target_lun})
            LOG.info(msg)
        else:
            self._map_lun(vol_instance, connector, target_portlist)
            mapdata = self._get_mapdata(vol_instance,
                                        connector, target_portlist)
        # end of if

        mapdata['target_discoverd'] = True
        mapdata['volume_id'] = volume['id']

        if self.protocol == 'fc':
            device_info = {'driver_volume_type': 'fibre_channel',
                           'data': mapdata}
        elif self.protocol == 'iSCSI':
            device_info = {'driver_volume_type': 'iscsi',
                           'data': mapdata}
        # end of if

        LOG.debug('initialize_connection,'
                  'device_info:%(info)s,'
                  'Exit method'
                  % {'info': device_info})
        return device_info

    def terminate_connection(self, volume, connector, force=False, **kwargs):
        '''
        Disallow connection from connector
        '''
        LOG.debug('terminate_connection,Enter method')

        self.conn = self._get_eternus_connection()
        self._unmap_lun(volume, connector)

        LOG.debug('terminate_connection,Exit method')
        return

    @eternus_dx_lockutils.FJDXLockutils('vol', 'cinder-', True)
    def extend_volume(self, volume, new_size):
        '''
        extend volume on ETERNUS
        '''
        LOG.debug('extend_volume,Enter method')

        # conversion of the unit. GB to B
        volumesize = new_size * units.Gi
        # create to volumename on ETERNUS from cinder VolumeID
        volumename = self._create_volume_name(volume['id'])
        # get source volume instance
        source_volume_instance = self._find_lun(volume)
        if source_volume_instance is None:
            msg = (_LE('extend_volume,'
                       'source_volume:%(volumename)s,'
                       'not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('extend_volume,volumename:%(volumename)s,'
                  'volumesize:%(volumesize)u,'
                  'source volume instance:%(source_volume_instance)s,'
                  % {'volumename': volumename,
                     'volumesize': volumesize,
                     'source_volume_instance':
                         str(source_volume_instance.path)})

        self.conn = self._get_eternus_connection()

        pool_name, pool = self._find_pool_from_volume(volume)
        if pool is None:
            msg = (_LE('extend_volume,'
                       'eternus_pool:%(eternus_pool)s,'
                       'not found.')
                   % {'eternus_pool': pool_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        # set pooltype
        if 'RSP' in pool['InstanceID']:
            pooltype = RAIDGROUP
        else:
            pooltype = TPPOOL
        # end of if

        configservice = self._find_eternus_service(STOR_CONF)
        if configservice is None:
            msg = (_LE('extend_volume,volume:%(volume)s,'
                       'volumename:%(volumename)s,'
                       'eternus_pool:%(eternus_pool)s,'
                       'Error!! Storage Configuration Service is None.')
                   % {'volume': volume,
                      'volumename': volumename,
                      'eternus_pool': pool_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if
        LOG.debug('extend_volume,CreateOrModifyElementFromStoragePool,'
                  'ConfigService:%(service)s,'
                  'ElementName:%(volumename)s,'
                  'InPool:%(eternus_pool)s,'
                  'ElementType:%(pooltype)u,'
                  'Size:%(volumesize)u,'
                  'TheElement:%(source_volume_instance)s'
                  % {'service': configservice,
                     'volumename': volumename,
                     'eternus_pool': pool_name,
                     'pooltype': pooltype,
                     'volumesize': volumesize,
                     'source_volume_instance':
                         str(source_volume_instance.path)})

        # Invoke method for extend volume
        rc, errordesc, job = self._exec_eternus_service(
            'CreateOrModifyElementFromStoragePool',
            configservice,
            ElementName=volumename,
            InPool=pool,
            ElementType=pywbem.Uint16(pooltype),
            Size=pywbem.Uint64(volumesize),
            TheElement=source_volume_instance.path)

        if rc != 0L:
            msg = (_LE('extend_volume,'
                       'volumename:%(volumename)s,'
                       'Return code:%(rc)lu,'
                       'Error:%(errordesc)s,'
                       'PoolType:%(pooltype)s')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc,
                      'pooltype': POOL_TYPE_dic[pooltype]})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('extend_volume,'
                  'volumename:%(volumename)s,'
                  'Return code:%(rc)lu,'
                  'Error:%(errordesc)s,'
                  'Pool Name:%(eternus_pool)s,'
                  'Pool Type:%(pooltype)s,'
                  'Exit method'
                  % {'volumename': volumename,
                     'rc': rc,
                     'errordesc': errordesc,
                     'eternus_pool': pool_name,
                     'pooltype': POOL_TYPE_dic[pooltype]})
        return

    @eternus_dx_lockutils.FJDXLockutils('update', 'cinder-', True)
    def update_volume_stats(self):
        '''
        get pool capacity.
        '''
        LOG.debug('update_volume_stats,Enter method')

        self.conn = self._get_eternus_connection()
        poolname_list = self._get_drvcfg('EternusPool', multiple=True)
        (pools, notfound_poolname) = self._find_pools(poolname_list)

        # Existence check the pool
        if notfound_poolname:
            msg = (_LI('update_volume_stats,'
                       'eternus_pool:%(notfound_poolname)s,not found.')
                   % {'notfound_poolname': notfound_poolname})
            LOG.info(msg)
        # end of if

        pools_info = self.scheduler.update_pools(pools)

        self.stats['total_capacity_gb'] = pools_info['total_capacity_gb']
        self.stats['free_capacity_gb'] = pools_info['free_capacity_gb']
        self.stats['multiattach'] = True

        LOG.debug('update_volume_stats,'
                  'eternus_pool:%(poolname_list)s,'
                  'total capacity[%(total)s],'
                  'free capacity[%(free)s],'
                  'Exit method'
                  % {'poolname_list': poolname_list,
                     'total': self.stats['total_capacity_gb'],
                     'free': self.stats['free_capacity_gb']})

        return self.stats

    def _find_pools(self, poolname_list):
        '''
        find pool instance information on ETERNUS
        '''
        LOG.debug('_find_pools,Enter method')

        target_poolname = list(poolname_list)

        try:
            tppoollist = self._enum_eternus_instances(
                'FUJITSU_ThinProvisioningPool')
            rgpoollist = self._enum_eternus_instances(
                'FUJITSU_RAIDStoragePool')
        except:
            msg = (_LE('_find_pools,'
                       'poolname_list:%(poolname_list)s,'
                       'EnumerateInstances,'
                       'cannot connect to ETERNUS.')
                   % {'poolname_list': poolname_list})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        pools_dic = {}
        for tppool in tppoollist:
            tppoolname = tppool['ElementName']
            if tppoolname in target_poolname:
                try:
                    volume_list = self._assoc_eternus_names(
                        tppool.path,
                        AssocClass='FUJITSU_AllocatedFromStoragePool',
                        ResultClass='FUJITSU_StorageVolume')
                    volume_count = len(volume_list)
                except:
                    msg = (_LE('_find_pools,'
                               'tppoolname:%(tppoolname)s,'
                               'Associator Names,'
                               'cannot connect to ETERNUS.')
                           % {'tppoolname': tppoolname})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                pool = self.scheduler.create_pool_info(tppool,
                                                       volume_count, 'TPP')
                target_poolname.remove(tppoolname)

                pool_idx = poolname_list.index(tppoolname)
                pools_dic[pool_idx] = pool
            # end of if
        # end of for tppoollist

        for rgpool in rgpoollist:
            rgpoolname = rgpool['ElementName']
            if rgpoolname in target_poolname:
                try:
                    volume_list = self._assoc_eternus_names(
                        rgpool.path,
                        AssocClass='FUJITSU_AllocatedFromStoragePool',
                        ResultClass='FUJITSU_StorageVolume')
                    volume_count = len(volume_list)
                except:
                    msg = (_LE('_find_pools,'
                               'rgpoolname:%(rgpoolname)s,'
                               'Associator Names,'
                               'cannot connect to ETERNUS.')
                           % {'rgpoolname': rgpoolname})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                pool = self.scheduler.create_pool_info(rgpool,
                                                       volume_count,
                                                       'RAID')
                target_poolname.remove(rgpoolname)

                pool_idx = poolname_list.index(rgpoolname)
                pools_dic[pool_idx] = pool
            # end of if
        # end of for rgpoollist

        temp_list = sorted(pools_dic.iteritems(), key=lambda i: i[0])
        pools = [i[1] for i in temp_list]

        if len(pools) == 0:
            msg = (_LW('_find_pools, all the EternusPools in driver'
                       ' configuration file are not exist. '
                       'Please edit driver configuration file'))
            LOG.warn(msg)

        LOG.debug('_find_pools,'
                  'pools: %(pools)s,'
                  'notfound_pools: %(notfound_pools)s,'
                  'Exit method.'
                  % {'pools': str(pools),
                     'notfound_pools': str(target_poolname)})

        return (pools, target_poolname)

    def _find_pool_from_volume(self, volume, vol_instance=None):
        '''
        find pool instance informaton on ETERNUS
        '''
        LOG.debug('_find_pool_from_volume,Enter method')

        pool_name = None

        v_metadata = volume.get('volume_metadata')
        for data in v_metadata:
            if data['key'] == "FJ_Pool_Name":
                pool_name = data['value']
                break
        # end of for v_metadata

        if pool_name is None:
            msg = (_LE('_find_pool_from_volume error'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        pool_instance = self._find_pool(pool_name)

        LOG.debug('_find_pool_from_volume,Exit method')
        return (pool_name, pool_instance)

    def _get_mapdata(self, vol_instance, connector, target_portlist):
        '''
        return mapping order
        '''
        LOG.debug('_get_mapdata,Enter method')

        mapdata = None

        # find affinity group
        # attach the connector and include the volume
        aglist = self._find_affinity_group(connector, vol_instance)
        if len(aglist) == 0:
            LOG.debug('_get_mapdata,ag_list:%s' % (aglist))
        else:
            if self.protocol == 'fc':
                mapdata = self._get_mapdata_fc(aglist, vol_instance,
                                               target_portlist)
            elif self.protocol == 'iSCSI':
                mapdata = self._get_mapdata_iscsi(aglist, vol_instance)

        LOG.debug('_get_mapdata,Exit method')
        return mapdata

    def _get_mapdata_fc(self, aglist, vol_instance, target_portlist):
        '''
        return mapping information
        '''
        LOG.debug('_get_mapdata_fc,Enter method')

        target_wwn = []

        try:
            ag_volmaplist = self._reference_eternus_names(
                aglist[0],
                ResultClass='CIM_ProtocolControllerForUnit')
            vo_volmaplist = self._reference_eternus_names(
                vol_instance.path,
                ResultClass='CIM_ProtocolControllerForUnit')
        except:
            msg = (_LE('_get_mapdata_fc,'
                       'affinitygroup:%(ag)s,'
                       'ReferenceNames,'
                       'cannot connect to ETERNUS.')
                   % {'ag': aglist[0]})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        volmap = None
        for vo_volmap in vo_volmaplist:
            if vo_volmap in ag_volmaplist:
                volmap = vo_volmap
                break
            # enf of if
        # end of for vo_volmaplist

        try:
            volmapinstance = self._get_eternus_instance(
                volmap,
                LocalOnly=False)
        except Exception:
            msg = (_LE('_get_mapdata_fc,'
                       'volmap:%(volmap)s,'
                       'GetInstance,'
                       'cannot connect to ETERNUS.')
                   % {'volmap': volmap})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        target_lun = int(volmapinstance['DeviceNumber'], 16)

        for target_port in target_portlist:
            target_wwn.append(target_port['Name'])
        # end of for target_portlist

        LOG.debug('_get_mapdata_fc,Exit method')
        mapdata = {'target_lun': target_lun,
                   'target_wwn': target_wwn}
        return mapdata

    def _get_mapdata_iscsi(self, aglist, vol_instance):
        '''
        return mapping information
        '''
        LOG.debug('_get_mapdata_iscsi,Enter method')

        target_portals = []
        target_iqns = []
        target_luns = []

        try:
            vo_volmaplist = self._reference_eternus_names(
                vol_instance.path,
                ResultClass='CIM_ProtocolControllerForUnit')
        except:
            msg = (_LE('_get_mapdata_iscsi,'
                       'vol_instance:%(vol_instance)s,'
                       'ReferenceNames,'
                       'cannot connect to ETERNUS.')
                   % {'vol_instance': vol_instance})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        target_properties_list = self._get_eternus_iscsi_properties()
        target_list = [prop[0] for prop in target_properties_list]
        properties_list = \
            [(prop[1], prop[2]) for prop in target_properties_list]

        for ag in aglist:
            try:
                iscsi_endpointlist = \
                    self._assoc_eternus_names(
                        ag,
                        AssocClass='FUJITSU_SAPAvailableForElement',
                        ResultClass='FUJITSU_iSCSIProtocolEndpoint')
            except:
                msg = (_LE('_get_mapdata_iscsi,'
                           'Associators,'
                           'cannot connect to ETERNUS.'))
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            iscsi_endpoint = iscsi_endpointlist[0]
            if iscsi_endpoint not in target_list:
                continue
            # end of if

            idx = target_list.index(iscsi_endpoint)
            target_portal, target_iqn = properties_list[idx]

            try:
                ag_volmaplist = self._reference_eternus_names(
                    ag,
                    ResultClass='CIM_ProtocolControllerForUnit')
            except:
                msg = (_LE('_get_mapdata_iscsi,'
                           'affinitygroup:%(ag)s,'
                           'ReferenceNames,'
                           'cannot connect to ETERNUS.')
                       % {'ag': ag})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            volmap = None
            for vo_volmap in vo_volmaplist:
                if vo_volmap in ag_volmaplist:
                    volmap = vo_volmap
                    break
                # end of if
            # end of for vo_volmaplist
            if volmap is None:
                continue
            # end of if

            try:
                volmapinstance = self._get_eternus_instance(
                    volmap,
                    LocalOnly=False)
            except Exception:
                msg = (_LE('_get_mapdata_iscsi,'
                           'volmap:%(volmap)s,'
                           'GetInstance,'
                           'cannot connect to ETERNUS.')
                       % {'volmap': volmap})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            target_lun = int(volmapinstance['DeviceNumber'], 16)

            target_portals.append(target_portal)
            target_iqns.append(target_iqn)
            target_luns.append(target_lun)
        # end of for aglist

        target_portal = target_portals[0]
        target_iqn = target_iqns[0]
        target_lun = target_luns[0]

        for i in range(len(target_portals)):
            ip = target_portals[i].split(':')[0]
            if self._is_target_alive(ip) is True:
                target_portal = target_portals[i]
                target_iqn = target_iqns[i]
                target_lun = target_luns[i]
                break
            # end of if
        else:
            LOG.warn(_LW('_get_mapdata_iscsi,'
                         'all target did not respond.'
                         'target_portals: %(target_portals)s')
                     % {'target_portals': target_portals})
        # end of for target_portals

        mapdata = {'target_portals': target_portals,
                   'target_iqns': target_iqns,
                   'target_luns': target_luns,
                   'target_lun': target_lun,
                   'target_portal': target_portal,
                   'target_iqn': target_iqn}

        LOG.debug('_get_mapdata_iscsi,Exit method')
        return mapdata

    def _get_drvcfg(self,
                    tagname, filename=None, multiple=False, allowNone=False):
        '''
        read from driver configuration file.
        '''
        LOG.debug('_get_drvcfg,Enter method')

        if filename is None:
            # set default configuration file name
            filename = self.configuration.cinder_eternus_config_file
        # end of if

        LOG.debug("_get_drvcfg input[%s][%s]" % (filename, tagname))

        tree = parse(filename)
        elem = tree.getroot()

        ret = None
        if multiple is False:
            ret = elem.findtext(".//" + tagname)
        else:
            ret = []
            for e in elem.findall(".//" + tagname):
                if (e.text is not None) and (e.text not in ret):
                    ret.append(e.text)
                # end of if
            # end of for elem.findall
        # end of if

        if ret is None or len(ret) == 0:
            if allowNone is False:
                msg = (_LE('_get_drvcfg,'
                           'filename:%(filename)s,'
                           'tagname:%(tagname)s,'
                           'data is None!!  '
                           'Please edit driver configuration file and correct'
                           ' an error. ')
                       % {'filename': filename,
                          'tagname': tagname})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                ret = None
            # end of if
        # end of if

        LOG.debug("_get_drvcfg output[%s],Exit method" % (ret))
        return ret

    def _get_eternus_connection(self, filename=None):
        '''
        return WBEM connection
        '''
        LOG.debug('_get_eternus_connection [%s],Enter method' % filename)

        ip = self._get_drvcfg('EternusIP', filename)
        port = self._get_drvcfg('EternusPort', filename)
        user = self._get_drvcfg('EternusUser', filename)
        passwd = self._get_drvcfg('EternusPassword', filename)
        url = 'http://' + ip + ':' + port

        conn = pywbem.WBEMConnection(url, (user, passwd),
                                     default_namespace='root/eternus')

        if conn is None:
            msg = (_LE('_get_eternus_connection,'
                       'filename:%(filename)s,'
                       'ip:%(ip)s,'
                       'port:%(port)s,'
                       'user:%(user)s,'
                       'passwd:****,'
                       'url:%(url)s,'
                       'FAILED!!.')
                   % {'filename': filename,
                      'ip': ip,
                      'port': port,
                      'user': user,
                      'url': url})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('_get_eternus_connection,[%s],Exit method' % (conn))
        return conn

    def _create_volume_name(self, id_code):
        '''
        create volume_name on ETERNUS from id on OpenStack.
        '''
        LOG.debug('_create_volume_name [%s],Enter method.' % id_code)

        if id_code is None:
            msg = (_LE('_create_volume_name,id_code is None.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        m = hashlib.md5()
        m.update(id_code)
        ret = VOL_PREFIX + str(base64.urlsafe_b64encode(m.digest()))

        LOG.debug('_create_volume_name,'
                  'ret:%(ret)s,'
                  'Exit method.'
                  % {'ret': ret})
        return ret

    def _get_snap_pool_name(self, snapshot):
        '''
        get pool name for SDV
        '''
        LOG.debug('_get_snap_pool_name,Enter method')

        snap_pool_name = self._get_drvcfg('EternusSnapPool', allowNone=True)
        if snap_pool_name is None:
            pool_instance = self._get_pool_from_scheduler(
                snapshot['volume_size'], use_tpp=False)
            snap_pool_name = pool_instance['name']
        # end of if

        LOG.debug('_get_snap_pool_name,Exit method')
        return snap_pool_name

    def _find_pool(self, eternus_pool, detail=False):
        '''
        find Instance or InstanceName of pool by pool name on ETERNUS.
        '''
        LOG.debug('_find_pool,Enter method')

        tppoollist = []
        rgpoollist = []

        # get pools info form CIM instance(include info about instance path)
        try:
            tppoollist = self._enum_eternus_instances(
                'FUJITSU_ThinProvisioningPool')
            rgpoollist = self._enum_eternus_instances(
                'FUJITSU_RAIDStoragePool')
        except Exception:
            msg = (_LE('_find_pool,'
                       'eternus_pool:%(eternus_pool)s,'
                       'EnumerateInstances,'
                       'cannot connect to ETERNUS.')
                   % {'eternus_pool': eternus_pool})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # make total pools list
        poollist = tppoollist + rgpoollist

        # one eternus backend has only one special pool name
        # so just use pool name can get the target pool
        for pool in poollist:
            if pool['ElementName'] == eternus_pool:
                poolinstance = pool
                break
            # end of if
        else:
            poolinstance = None
        # end of for poollist

        if poolinstance is None:
            ret = None
        elif detail is True:
            ret = poolinstance
        else:
            ret = poolinstance.path
        # end of if

        LOG.debug('_find_pool,'
                  'pool: %(ret)s,'
                  'Exit method.'
                  % {'ret': ret})
        return ret

    def _find_eternus_service(self, classname):
        '''
        find CIM instance
        '''
        LOG.debug('_find_eternus_service,'
                  'classname:%(a)s,'
                  'Enter method'
                  % {'a': str(classname)})

        try:
            services = self._enum_eternus_instance_names(
                str(classname))
        except:
            msg = (_LE('_find_eternus_service,'
                       'classname:%(classname)s,'
                       'EnumerateInstanceNames,'
                       'cannot connect to ETERNUS.')
                   % {'classname': str(classname)})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        ret = services[0]
        LOG.debug('_find_eternus_service,'
                  'classname:%(classname)s,'
                  'ret:%(ret)s,'
                  'Exit method'
                  % {'classname': classname,
                     'ret': (str(ret))})
        return ret

    @eternus_dx_lockutils.FJDXLockutils('SMIS-exec', 'cinder-', True)
    def _exec_eternus_service(self, classname, instanceNameList, **param_dict):
        '''
        Execute SMI-S Method
        '''
        LOG.debug('_exec_eternus_service,'
                  'classname:%(a)s,'
                  'instanceNameList:%(b)s,'
                  'parameters:%(c)s,'
                  'Enter method'
                  % {'a': str(classname),
                     'b': str(instanceNameList),
                     'c': str(param_dict)})

        # use InvokeMethod
        try:
            rc, retdata = self.conn.InvokeMethod(
                classname,
                instanceNameList,
                **param_dict)
        except Exception:
            if rc is None:
                msg = (_LE('_exec_eternus_service,'
                           'classname:%(classname)s,'
                           'InvokeMethod,'
                           'cannot connect to ETERNUS.')
                       % {'classname': str(classname)})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if
        # convert errorcode to error description

        # If the result has job information, wait for job complete
        if "Job" in retdata:
            rc = self._wait_for_job_complete(self.conn, retdata)
        # end of if

        errordesc = RETCODE_dic.get(six.text_type(rc), UNDEF_MSG)

        ret = (rc, errordesc, retdata)

        LOG.debug('_exec_eternus_service,'
                  'classname:%(a)s,'
                  'instanceNameList:%(b)s,'
                  'parameters:%(c)s,'
                  'Return code:%(rc)s,'
                  'Error:%(errordesc)s,'
                  'Retrun data:%(retdata)s,'
                  'Exit method'
                  % {'a': str(classname),
                     'b': str(instanceNameList),
                     'c': str(param_dict),
                     'rc': str(rc),
                     'errordesc': errordesc,
                     'retdata': str(retdata)})
        return ret

    @eternus_dx_lockutils.FJDXLockutils('SMIS-other', 'cinder-', True)
    def _enum_eternus_instances(self, classname, retry=20, retry_interval=5):
        '''
        Enumerate Instances
        '''
        LOG.debug('_enum_eternus_instances,Enter method')

        for retry_num in range(retry):
            try:
                ret = self.conn.EnumerateInstances(classname)
                break
            except Exception as e:
                LOG.info(_LI('_enum_eternus_instances,'
                             ' reason:%(reason)s,try (%(retrynum)s)')
                         % {'reason': str(e.args),
                            'retrynum': str(retry_num + 1)})
                time.sleep(retry_interval)
                continue
        else:
            msg = ('_enum_eternus_instances, Error')
            raise exception.VolumeBackendAPIException(data=msg)
        # end of for rety

        LOG.debug('_enum_eternus_instances,Exit method')
        return ret

    @eternus_dx_lockutils.FJDXLockutils('SMIS-other', 'cinder-', True)
    def _enum_eternus_instance_names(self,
                                     classname,
                                     conn=None, retry=20, retry_interval=5):
        '''
        Enumerate Instance Names
        '''
        LOG.debug('_enum_eternus_instance_names,Enter method')

        if conn is None:
            conn = self.conn
        # end of if

        for retry_num in range(retry):
            try:
                ret = conn.EnumerateInstanceNames(classname)
                break
            except Exception as e:
                LOG.info(_LI('_enum_eternus_instance_names,'
                             ' reason:%(reason)s,try (%(retrynum)s)')
                         % {'reason': str(e.args),
                            'retrynum': str(retry_num + 1)})
                time.sleep(retry_interval)
                continue
        else:
            msg = ('_enum_eternus_instance_names, Error')
            raise exception.VolumeBackendAPIException(data=msg)
        # end of for rety

        LOG.debug('_enum_eternus_instance_names,Exit method')
        return ret

    @eternus_dx_lockutils.FJDXLockutils('SMIS-getinstance', 'cinder-', True)
    def _get_eternus_instance(self, classname, conn=None, AllowNone=False,
                              retry=20, retry_interval=5, **param_dict):
        '''
        Get Instance
        '''
        LOG.debug('_get_eternus_instance,Enter method')

        if conn is None:
            conn = self.conn
        # end of if

        ret = None
        for retry_num in range(retry):
            try:
                ret = conn.GetInstance(classname, **param_dict)
                break
            except Exception as e:
                if (e.args[0] == 6) and (AllowNone is True):
                    break
                else:
                    LOG.info(_LI('_get_eternus_instance,'
                                 ' reason:%(reason)s,try (%(retrynum)s)')
                             % {'reason': str(e.args),
                                'retrynum': str(retry_num + 1)})
                    time.sleep(retry_interval)
                    continue
                # end of if
        else:
            msg = ('_get_eternus_instance, Error')
            raise exception.VolumeBackendAPIException(data=msg)
        # end of for rety

        LOG.debug('_get_eternus_instance,Exit method')
        return ret

    @eternus_dx_lockutils.FJDXLockutils('SMIS-other', 'cinder-', True)
    def _assoc_eternus(self,
                       classname, retry=20, retry_interval=5, **param_dict):
        '''
        Associator
        '''
        LOG.debug('_assoc_eternus,Enter method')

        for retry_num in range(retry):
            try:
                ret = self.conn.Associators(classname, **param_dict)
                break
            except Exception as e:
                LOG.info(_LI('_assoc_eternus,'
                             ' reason:%(reason)s,retry (%(retrynum)s)')
                         % {'reason': str(e.args),
                            'retrynum': str(retry_num + 1)})
                time.sleep(retry_interval)
                continue
        else:
            msg = ('_assoc_eternus, Error')
            raise exception.VolumeBackendAPIException(data=msg)
        # end of for rety

        LOG.debug('_assoc_eternus,Exit method')
        return ret

    @eternus_dx_lockutils.FJDXLockutils('SMIS-other', 'cinder-', True)
    def _assoc_eternus_names(self,
                             classname,
                             retry=20, retry_interval=5, **param_dict):
        '''
        Associator Names
        '''
        LOG.debug('_assoc_eternus_names,Enter method')

        for retry_num in range(retry):
            try:
                ret = self.conn.AssociatorNames(classname, **param_dict)
                break
            except Exception as e:
                LOG.info(_LI('_assoc_eternus_names,'
                             ' reason:%(reason)s,try (%(retrynum)s)')
                         % {'reason': str(e.args),
                            'retrynum': str(retry_num + 1)})
                time.sleep(retry_interval)
                continue
        else:
            msg = ('_assoc_eternus_names, Error')
            raise exception.VolumeBackendAPIException(data=msg)
        # end of for rety

        LOG.debug('_assoc_eternus_names,Exit method')
        return ret

    @eternus_dx_lockutils.FJDXLockutils('SMIS-other', 'cinder-', True)
    def _reference_eternus_names(self,
                                 classname,
                                 retry=20, retry_interval=5, **param_dict):
        '''
        Refference Names
        '''
        LOG.debug('_reference_eternus_names,Enter method')

        for retry_num in range(retry):
            try:
                ret = self.conn.ReferenceNames(classname, **param_dict)
                break
            except Exception as e:
                LOG.info(_LI('_reference_eternus_names,'
                             ' reason:%(reason)s,try (%(retrynum)s)')
                         % {'reason': str(e.args),
                            'retrynum': str(retry_num + 1)})
                time.sleep(retry_interval)
                continue
        else:
            msg = ('_reference_eternus_names, Error')
            raise exception.VolumeBackendAPIException(data=msg)
        # end of for rety

        LOG.debug('_reference_eternus_names,Exit method')
        return ret

    def _create_eternus_instance_name(self, classname, bindings):
        '''
        create volume instance name from metadata
        '''
        LOG.debug('_create_eternus_instance_name,Enter method')

        instancename = None

        try:
            instancename = pywbem.CIMInstanceName(
                classname,
                namespace='root/eternus',
                keybindings=bindings)
        except NameError:
            instancename = None

        LOG.debug('_create_eternus_instance_name,Exit method')
        return instancename

    def _find_lun(self, volume, use_service_name=False):
        '''
        find lun instance from volume class or volumename on ETERNUS.
        '''
        LOG.debug('_find_lun,Enter method')

        volumeinstance = None

        volumename = self._create_volume_name(volume['id'])

        if use_service_name is False:
            conn = self.conn
        else:
            service_name = volume['host'].split('@', 1)[1].split('#')[0]
            conf_filename = \
                Configuration(
                    FJ_ETERNUS_DX_OPT_list,
                    config_group=service_name).cinder_eternus_config_file
            conn = self._get_eternus_connection(conf_filename)
        # end of if

        try:
            location = eval(volume['provider_location'])
            classname = location['classname']
            bindings = location['keybindings']

            if (classname is not None) and (bindings is not None):
                LOG.debug('_find_lun,'
                          'classname:%(classname)s,'
                          'bindings:%(bindings)s'
                          % {'classname': classname,
                             'bindings': bindings})
                volume_instance_name = \
                    self._create_eternus_instance_name(classname, bindings)

                LOG.debug('_find_lun,'
                          'volume_insatnce_name:%(volume_instance_name)s'
                          % {'volume_instance_name': volume_instance_name})

                vol_instance = \
                    self._get_eternus_instance(volume_instance_name,
                                               conn=conn,
                                               AllowNone=True)

                if vol_instance['ElementName'] == volumename:
                    volumeinstance = vol_instance
                # end of if
            # end of if
        except:
            volumeinstance = None
            LOG.debug('_find_lun,'
                      'cannot get volume instance from provider location, '
                      'search all volume using EnumerateInstanceNames')

        if volumeinstance is None:
            # for old version

            LOG.debug('_find_lun,'
                      'volumename:%(volumename)s'
                      % {'volumename': volumename})

            # get volume instance from volumename on ETERNUS
            try:
                namelist = self._enum_eternus_instance_names(
                    'FUJITSU_StorageVolume')
            except:
                msg = (_LE('_find_lun,'
                           'volumename:%(volumename)s,'
                           'EnumerateInstanceNames,'
                           'cannot connect to ETERNUS.')
                       % {'volumename': volumename})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            for name in namelist:
                try:
                    vol_instance = self._get_eternus_instance(
                        name, conn=conn, AllowNone=True)

                    if vol_instance['ElementName'] == volumename:
                        volumeinstance = vol_instance

                        LOG.debug('_find_lun,'
                                  'volumename:%(volumename)s,'
                                  'vol_instance:%(vol_instance)s.'
                                  % {'volumename': volumename,
                                     'vol_instance': str(volumeinstance.path)})
                        break
                    # end of if
                except:
                    vol_instance = None
            else:
                LOG.debug('_find_lun,'
                          'volumename:%(volumename)s,'
                          'volume not found on ETERNUS.'
                          % {'volumename': volumename})
            # end of for namelist
        # end of if

        LOG.debug('_find_lun,Exit method')
        return volumeinstance

    def _find_copysession(self, vol_instance):
        '''
        find copysession from volumename on ETERNUS
        '''
        LOG.debug('_find_copysession, Enter method')

        try:
            cpsessionlist = self.conn.ReferenceNames(
                vol_instance.path,
                ResultClass='FUJITSU_StorageSynchronized')
        except Exception:
            msg = (_LE('_find_copysession,'
                       'ReferenceNames,'
                       'vol_instance: %(vol_instance_path)s,'
                       'Cannot connect to ETERNUS')
                   % {'vol_instance_path': str(vol_instance.path)})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('_find_copysession,'
                  'cpsessionlist: %(cpsessionlist)s'
                  % {'cpsessionlist': cpsessionlist})

        LOG.debug('_find_copysession,Exit method')
        return cpsessionlist

    def _wait_for_copy_complete(self, cpsession):
        '''
        Wait for the completion of copy session
        '''
        LOG.debug('_wait_for_copy_complete,Enter method')

        cpsession_instance = None

        while True:
            try:
                cpsession_instance = self.conn.GetInstance(
                    cpsession,
                    LocalOnly=False)
            except Exception:
                cpsession_instance = None

            # if copy session is none,
            # it means copy session was finished,break and return
            if cpsession_instance is None:
                break
            # end of if

            LOG.debug('_wait_for_copy_complete,'
                      'find target copysession,'
                      'wait for end of copysession')

            if cpsession_instance['CopyState'] == BROKEN:
                msg = (_LE('_wait_for_copy_complete,'
                           'cpsession:%(cpsession)s,'
                           'copysession state is BROKEN')
                       % {'cpsession': cpsession})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            time.sleep(10)
        # end of while

        LOG.debug('_wait_for_copy_complete,Exit method')
        return

    def _delete_copysession(self, cpsession):
        '''
        delete copysession
        '''
        LOG.debug('_delete_copysession,Enter method.[%s]' % cpsession)

        try:
            snapshot_instance = self._get_eternus_instance(
                cpsession,
                LocalOnly=False,
                AllowNone=True)
        except:
            msg = (_LE('_delete_copysession,'
                       'copysession:%(cpsession)s,'
                       'GetInstance,'
                       'cannot connect to ETERNUS.')
                   % {'cpsession': cpsession})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if snapshot_instance is None:
            LOG.info(_LI('_delete_copysession,'
                         'The copysession was already completed.'))
        else:
            copytype = snapshot_instance['CopyType']

            # set oparation code
            # SnapOPC: 19 (Return To ResourcePool)
            # OPC:8 (Detach)
            # EC/REC:8 (Detach)
            operation = OPERATION_dic.get(copytype, None)
            if operation is None:
                msg = (_LE('_delete_copysession,'
                           'copy session type is undefined!'
                           'Copy session:%(cpsession)s;'
                           'copy type:%(copytype)s')
                       % {'cpsession': cpsession,
                          'copytype': copytype})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            repservice = self._find_eternus_service(REPL)
            if repservice is None:
                msg = (_LE('_delete_copysession,'
                           'Cannot find Replication Service to '
                           'delete copysession'))
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            # Invoke method for delete copysession
            rc, errordesc, job = self._exec_eternus_service(
                'ModifyReplicaSynchronization',
                repservice,
                Operation=pywbem.Uint16(operation),
                Synchronization=cpsession,
                Force=True,
                WaitForCopyState=pywbem.Uint16(15))

            LOG.debug('_delete_copysession,'
                      'copysession:%(cpsession)s,'
                      'operation:%(operation)s,'
                      'Return code:%(rc)lu,'
                      'Error:%(errordesc)s,'
                      'Exit method'
                      % {'cpsession': cpsession,
                         'operation': operation,
                         'rc': rc,
                         'errordesc': errordesc})
            if rc == COPYSESSION_NOT_EXIST:
                LOG.debug('_delete_copysession, '
                          'cpsession: %(cpsession)s, '
                          'copysession is not exist.'
                          % {'cpsession': cpsession})
            elif rc != 0L:
                msg = (_LE('_delete_copysession,'
                           'copysession:%(cpsession)s,'
                           'operation:%(operation)s,'
                           'Return code:%(rc)lu,'
                           'Error:%(errordesc)s')
                       % {'cpsession': cpsession,
                          'operation': operation,
                          'rc': rc,
                          'errordesc': errordesc})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if
        # end of if

        LOG.debug('_delete_copysession,Exit method')
        return

    def _get_target_port(self):
        '''
        return target_portid
        '''
        LOG.debug('_get_target_port,Enter method')

        target_portlist = []

        if self.protocol == 'fc':
            prtcl_endpoint = 'FUJITSU_SCSIProtocolEndpoint'
            connection_type = 2
        elif self.protocol == 'iSCSI':
            prtcl_endpoint = 'FUJITSU_iSCSIProtocolEndpoint'
            connection_type = 7

        try:
            tgtportlist = self._enum_eternus_instances(prtcl_endpoint)
        except:
            msg = (_LE('_get_target_port,'
                       'EnumerateInstances,'
                       'cannot connect to ETERNUS.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for tgtport in tgtportlist:
            if tgtport['ConnectionType'] != connection_type:
                continue

            if (tgtport['RAMode'] & 0x7B) != 0x00:
                continue

            if 'SCGroupNo' in tgtport:
                continue

            target_portlist.append(tgtport)

            LOG.debug('_get_target_port,'
                      'connection type:%(cont)s,'
                      'ramode:%(ramode)s'
                      % {'cont': tgtport['ConnectionType'],
                         'ramode': tgtport['RAMode']})
        # end of for tgtportlist

        LOG.debug('_get_target_port,'
                  'target port: %(target_portid)s '
                  % {'target_portid': target_portlist})
        # end of if

        if len(target_portlist) == 0:
            msg = (_LE('_get_target_port,'
                       'protcol:%(protocol)s,'
                       'target_port does not found.')
                   % {'protocol': self.protocol})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('_get_target_port,Exit method')
        return target_portlist

    @eternus_dx_lockutils.FJDXLockutils('connect', 'cinder-', True)
    def _map_lun(self, vol_instance, connector, targetlist=None):
        '''map volume to host.'''

        volumename = vol_instance['ElementName']
        LOG.debug('_map_lun,'
                  'volume:%(volume)s,'
                  'connector:%(con)s,'
                  'Enter method'
                  % {'volume': volumename,
                     'con': connector})

        volume_uid = vol_instance['Name']
        initiatorlist = self._find_initiator_names(connector)
        aglist = self._find_affinity_group(connector)
        configservice = self._find_eternus_service(CTRL_CONF)

        if targetlist is None:
            targetlist = self._get_target_port()
        # end of if

        if configservice is None:
            msg = (_LE('_map_lun,'
                       'vol_instance.path:%(vol)s,'
                       'volumename:%(volumename)s,'
                       'volume_uid:%(uid)s,'
                       'initiator:%(initiator)s,'
                       'target:%(tgt)s'
                       'aglist:%(aglist)s'
                       'Cannot find Controller Configuration')
                   % {'vol': str(vol_instance.path),
                      'volumename': volumename,
                      'uid': volume_uid,
                      'initiator': initiatorlist,
                      'tgt': targetlist,
                      'aglist': aglist})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if
        LOG.debug('_map_lun,'
                  'vol_instance.path:%(vol)s,'
                  'volumename:%(volumename)s,'
                  'initiator:%(initiator)s,'
                  'target:%(tgt)s'
                  % {'vol': str(vol_instance.path),
                     'volumename': [volumename],
                     'initiator': initiatorlist,
                     'tgt': targetlist})

        if len(aglist) == 0:
            # create new affinity group
            for target in targetlist:
                LOG.debug('_map_lun,'
                          'lun_name:%(volume_uid)s,'
                          'Initiator:%(initiator)s,'
                          'target:%(tgt)s'
                          % {'volume_uid': [volume_uid],
                             'initiator': initiatorlist,
                             'tgt': target['Name']})

                rc, errordesc, job = self._exec_eternus_service(
                    'ExposePaths',
                    configservice,
                    LUNames=[volume_uid],
                    InitiatorPortIDs=initiatorlist,
                    TargetPortIDs=[target['Name']],
                    DeviceAccesses=[pywbem.Uint16(2)])

                LOG.debug('_map_lun,'
                          'Error:%(errordesc)s,'
                          'Return code:%(rc)lu,'
                          'Create new affinitygroup'
                          % {'errordesc': errordesc,
                             'rc': rc})

                if rc != 0L and rc != LUNAME_IN_USE:
                    LOG.warn(_LW('_map_lun,'
                                 'lun_name:%(volume_uid)s,'
                                 'Initiator:%(initiator)s,'
                                 'target:%(tgt)s,'
                                 'Return code:%(rc)lu,'
                                 'Error:%(errordesc)s')
                             % {'volume_uid': [volume_uid],
                                'initiator': initiatorlist,
                                'tgt': target['Name'],
                                'rc': rc,
                                'errordesc': errordesc})
                # end of if
            # end of for targetlist
        else:
            # add lun to affinity group
            for ag in aglist:
                LOG.debug('_map_lun,'
                          'ag:%(ag)s,'
                          'lun_name:%(volume_uid)s'
                          % {'ag': str(ag),
                             'volume_uid': volume_uid})

                rc, errordesc, job = self._exec_eternus_service(
                    'ExposePaths',
                    configservice, LUNames=[volume_uid],
                    DeviceAccesses=[pywbem.Uint16(2)],
                    ProtocolControllers=[ag])

                LOG.debug('_map_lun,'
                          'Error:%(errordesc)s,'
                          'Return code:%(rc)lu,'
                          'Add lun affinitygroup'
                          % {'errordesc': errordesc,
                             'rc': rc})

                if rc != 0L and rc != LUNAME_IN_USE:
                    LOG.warn(_LW('_map_lun,'
                                 'lun_name:%(volume_uid)s,'
                                 'Initiator:%(initiator)s,'
                                 'ag:%(ag)s,'
                                 'Return code:%(rc)lu,'
                                 'Error:%(errordesc)s')
                             % {'volume_uid': [volume_uid],
                                'initiator': initiatorlist,
                                'ag': ag,
                                'rc': rc,
                                'errordesc': errordesc})
                # end of if
            # end of for aglist
        # end of if
        LOG.debug('_map_lun,'
                  'volumename:%(volumename)s,'
                  'Exit method'
                  % {'volumename': volumename})
        return

    def _find_initiator_names(self, connector):
        '''
        return initiator names
        '''
        LOG.debug('_find_initiator_names,Enter method')

        initiatornamelist = []

        if self.protocol == 'fc' and connector['wwpns'] is not None:
            initiatornamelist = connector['wwpns']
        elif self.protocol == 'iSCSI' and connector['initiator'] is not None:
            initiatornamelist.append(connector['initiator'])
        # end of if

        if len(initiatornamelist) == 0:
            msg = (_LE('_find_initiator_names,'
                       'connector:%(connector)s,'
                       'not found initiator')
                   % {'connector': connector})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('_find_initiator_names,'
                  'initiator:%(initiator)s.'
                  'Exit method'
                  % {'initiator': initiatornamelist})

        return initiatornamelist

    def _find_affinity_group(self, connector, vol_instance=None):
        '''
        find affinity group from connector
        '''
        LOG.debug('_find_affinity_group,Enter method')

        affinity_grouplist = []

        initiatorlist = self._find_initiator_names(connector)

        if vol_instance is None:
            try:
                aglist = self._enum_eternus_instance_names(
                    'FUJITSU_AffinityGroupController')
            except:
                msg = (_LE('_find_affinity_group,'
                           'connector:%(connector)s,'
                           'EnumerateInstanceNames,'
                           'cannot connect to ETERNUS.')
                       % {'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_find_affinity_group,'
                      'affinity_groups:%s'
                      % aglist)
        else:
            try:
                aglist = self._assoc_eternus_names(
                    vol_instance.path,
                    AssocClass='FUJITSU_ProtocolControllerForUnit',
                    ResultClass='FUJITSU_AffinityGroupController')
            except:
                msg = (_LE('_find_affinity_group,'
                           'connector:%(connector)s,'
                           'AssociatorNames,'
                           'cannot connect to ETERNUS.')
                       % {'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_find_affinity_group,'
                      'vol_instance.path:%(vol)s,'
                      'affinity_groups:%(aglist)s'
                      % {'vol': vol_instance.path,
                         'aglist': aglist})
        # end of if
        for ag in aglist:
            try:
                hostaglist = self._assoc_eternus(
                    ag,
                    AssocClass='FUJITSU_AuthorizedTarget',
                    ResultClass='FUJITSU_AuthorizedPrivilege')
            except:
                msg = (_LE('_find_affinity_group,'
                           'connector:%(connector)s,'
                           'Associators,'
                           'cannot connect to ETERNUS.')
                       % {'connector': connector})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            for hostag in hostaglist:
                for initiator in initiatorlist:
                    if initiator.lower() not in hostag['InstanceID'].lower():
                        continue
                    # end of if
                    LOG.debug('_find_affinity_group,'
                              'AffinityGroup:%(ag)s'
                              % {'ag': ag})
                    affinity_grouplist.append(ag)
                    break
                # end of for initiatorlist
                break
            # end of for hostaglist
        # end of for aglist

        LOG.debug('_find_affinity_group,'
                  'initiators:%(initiator)s,'
                  'affinity_group:%(affinity_group)s.'
                  'Exit method'
                  % {'initiator': initiatorlist,
                     'affinity_group': affinity_grouplist})
        return affinity_grouplist

    @eternus_dx_lockutils.FJDXLockutils('connect', 'cinder-', True)
    def _unmap_lun(self, volume, connector, force=False):
        '''
        unmap volume from host
        '''
        LOG.debug('_unmap_lun,Enter method')

        volumename = self._create_volume_name(volume['id'])
        vol_instance = self._find_lun(volume)
        if vol_instance is None:
            LOG.info(_LI('_unmap_lun,'
                         'volumename:%(volumename)s,'
                         'volume not found.'
                         'Exit method')
                     % {'volumename': volumename})
            return
        # end of if

        volume_uid = vol_instance['Name']
        configservice = self._find_eternus_service(CTRL_CONF)

        if force is False:
            aglist = self._find_affinity_group(connector, vol_instance)
            if len(aglist) == 0:
                LOG.info(_LI('_unmap_lun,'
                             'volumename:%(volumename)s,'
                             'volume is not mapped.'
                             'Exit method')
                         % {'volumename': volumename})
                return
            # end of if
        else:
            try:
                aglist = self._assoc_eternus_names(
                    vol_instance.path,
                    AssocClass='CIM_ProtocolControllerForUnit',
                    ResultClass='FUJITSU_AffinityGroupController')
            except:
                msg = (_LE('_unmap_lun,'
                           'vol_instance.path:%(vol)s,'
                           'AssociatorNames,'
                           'cannot connect to ETERNUS.')
                       % {'vol': vol_instance.path})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('_unmap_lun,'
                      'vol_instance.path:%(vol)s,'
                      'affinity_groups:%(aglist)s'
                      % {'vol': vol_instance.path,
                         'aglist': aglist})
        # end of if

        if configservice is None:
            msg = (_LE('_unmap_lun,'
                       'vol_instance.path:%(vol)s,'
                       'volumename:%(volumename)s,'
                       'volume_uid:%(uid)s,'
                       'aglist:%(aglist)s'
                       'Cannot find Controller Configuration')
                   % {'vol': str(vol_instance.path),
                      'volumename': [volumename],
                      'uid': [volume_uid],
                      'aglist': aglist})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        for ag in aglist:
            LOG.debug('_unmap_lun,'
                      'volumename:%(volumename)s,'
                      'volume_uid:%(volume_uid)s,'
                      'AffinityGroup:%(ag)s'
                      % {'volumename': volumename,
                         'volume_uid': volume_uid,
                         'ag': ag})

            rc, errordesc, job = self._exec_eternus_service(
                'HidePaths',
                configservice,
                LUNames=[volume_uid],
                ProtocolControllers=[ag])

            LOG.debug('_unmap_lun,'
                      'Error:%(errordesc)s,'
                      'Return code:%(rc)lu'
                      % {'errordesc': errordesc,
                         'rc': rc})

            if rc == LUNAME_NOT_EXIST:
                LOG.debug('_unmap_lun,'
                          'volumename:%(volumename)s,'
                          'Invalid LUNames'
                          % {'volumename': volumename})
            elif rc != 0L:
                msg = (_LE('_unmap_lun,'
                           'volumename:%(volumename)s,'
                           'volume_uid:%(volume_uid)s,'
                           'AffinityGroup:%(ag)s,'
                           'Return code:%(rc)lu,'
                           'Error:%(errordesc)s')
                       % {'volumename': volumename,
                          'volume_uid': volume_uid,
                          'ag': ag,
                          'rc': rc,
                          'errordesc': errordesc})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if
        # end of for aglist
        LOG.debug('_unmap_lun,'
                  'volumename:%(volumename)s,'
                  'Exit method'
                  % {'volumename': volumename})
        return

    def _get_eternus_iscsi_properties(self):
        '''
        get target port iqns and target_portals
        '''
        LOG.debug('_get_eternus_iscsi_properties,Enter method')

        iscsi_properties_list = []

        iscsiip_list = self._get_drvcfg('EternusISCSIIP', multiple=True)
        iscsi_port = self.configuration.iscsi_port

        try:
            ip_endpointlist = self._enum_eternus_instance_names(
                'FUJITSU_IPProtocolEndpoint')
        except:
            msg = (_LE('_get_eternus_iscsi_properties,'
                       'iscsiip:%(iscsiip)s,'
                       'EnumerateInstanceNames,'
                       'cannot connect to ETERNUS.')
                   % {'iscsiip': iscsiip_list})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for ip_endpoint in ip_endpointlist:
            try:
                ip_endpoint_instance = self._get_eternus_instance(
                    ip_endpoint)
                ip_address = ip_endpoint_instance['IPv4Address']
                LOG.debug('_get_eternus_iscsi_properties,'
                          'instanceip:%(ip)s,'
                          'iscsiip:%(iscsiip)s'
                          % {'ip': ip_address,
                             'iscsiip': str(iscsiip_list)})
            except:
                msg = (_LE('_get_eternus_iscsi_properties,'
                           'iscsiip:%(iscsiip)s,'
                           'GetInstance,'
                           'cannot connect to ETERNUS.')
                       % {'iscsiip': str(iscsiip_list)})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if ip_address not in iscsiip_list:
                continue
            # end of if

            LOG.debug('_get_eternus_iscsi_properties,'
                      'find iscsiip: %(ip)s.' % {'ip': ip_address})
            try:
                tcp_endpointlist = self._assoc_eternus_names(
                    ip_endpoint,
                    AssocClass='CIM_BindsTo',
                    ResultClass='FUJITSU_TCPProtocolEndpoint')
            except:
                msg = (_LE('_get_eternus_iscsi_properties,'
                           'iscsiip:%(iscsiip)s,'
                           'AssociatorNames,'
                           'cannot connect to ETERNUS.')
                       % {'iscsiip': ip_address})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            for tcp_endpoint in tcp_endpointlist:
                try:
                    iscsi_endpointlist = \
                        self._assoc_eternus(tcp_endpoint,
                                            AssocClass='CIM_BindsTo',
                                            ResultClass='FUJITSU_iSCSI'
                                            'ProtocolEndpoint')
                except:
                    msg = (_LE('_get_eternus_iscsi_properties,'
                               'iscsiip:%(iscsiip)s,'
                               'AssociatorNames,'
                               'cannot connect to ETERNUS.')
                           % {'iscsiip': ip_address})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                for iscsi_endpoint in iscsi_endpointlist:
                    target_portal = "%s:%s" % (ip_address, iscsi_port)
                    iqn = iscsi_endpoint['Name'].split(',')[0]
                    iscsi_properties_list.append((iscsi_endpoint.path,
                                                  target_portal,
                                                  iqn))
                    LOG.debug('_get_eternus_iscsi_properties,'
                              'target_portal: %(target_portal)s,'
                              'iqn: %(iqn)s.'
                              % {'target_portal': target_portal,
                                 'iqn': iqn})
                # end of for iscsi_endpointlist
            # end of for tcp_endpointlist
        # end of for ip_endpointlist

        if len(iscsi_properties_list) == 0:
            msg = (_LE('_get_eternus_iscsi_properties,'
                       'iscsiip:%(iscsiip)s,'
                       'not found iqn')
                   % {'iscsiip': iscsiip_list})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        LOG.debug('_get_eternus_iscsi_properties,'
                  'iscsi_properties_list: %(iscsi_properties_list)s'
                  % {'iscsi_properties_list': iscsi_properties_list})

        LOG.debug('_get_eternus_iscsi_properties,Exit method')

        return iscsi_properties_list

    def _is_target_alive(self, ip):
        '''
        confirm whether target is alive or not
        '''
        LOG.debug('_is_target_alive,Enter method')

        ret = True

        for i in range(3):
            try:
                (out, _err) = utils.execute('ping', '-c', '1', ip)
                break
            except processutils.ProcessExecutionError:
                continue
        else:
            LOG.warn(_LW('_is_target_alive,'
                         ' target(%(ip)s) did not respond to icmp')
                     % {'ip': ip})
            ret = False
        # end of for range(3)

        LOG.debug('_is_target_alive,Exit method')
        return ret

    def _get_extra_specs(self, volume, key=None, default=None):
        '''
        get extra specs information from volume information
        '''
        LOG.debug('_get_extra_specs,Enter method')

        extra_specs = {}
        ret = default

        volume_type_id = volume.get('volume_type_id')

        if volume_type_id is not None:
            ctxt = context.get_admin_context()
            volume_type = volume_types.get_volume_type(ctxt, volume_type_id)
            extra_specs = volume_type.get('extra_specs')
        # end of if

        if extra_specs:
            if key is None:
                ret = extra_specs
            else:
                ret = extra_specs.get(key, default)
            # end of if
        # end of if

        LOG.debug('_get_extra_specs,Exit method')
        return ret

    def _get_qos_specs(self, volume):
        '''
        get qos specs information from volume information
        '''
        LOG.debug('_get_qos_specs,Enter method')

        qos_specs_dict = {}
        qos_specs_id = None

        volume_type_id = volume.get('volume_type_id')

        if volume_type_id is not None:
            ctxt = context.get_admin_context()
            volume_type = volume_types.get_volume_type(ctxt, volume_type_id)
            qos_specs_id = volume_type.get('qos_specs_id')
        # end of if

        if qos_specs_id is not None:
            qos_specs_dict = qos_specs.get_qos_specs(ctxt,
                                                     qos_specs_id)['specs']
        # end of if

        LOG.debug('_get_qos_specs,Exit method')
        return qos_specs_dict

    def _get_metadata(self, volume):
        '''
        get metadata using volume information and keys
        '''
        LOG.debug('_get_metadata,Enter method')

        ret = {}

        metadata = volume.get('volume_metadata', [])
        if metadata is None:
            metadata = []
        # end of if

        for data in metadata:
            ret[data['key']] = data['value']
        # end of for metadata

        LOG.debug('_get_metadata,Exit method')
        return ret

    def _get_bool(self, value):
        '''
        get True/False from String 'True'/'False'
        '''
        LOG.debug('_get_bool,Enter method')
        ret = False

        if str(value).lower() == "true":
            ret = True
        # end of if

        LOG.debug('_get_bool,Exit method')
        return ret

    def _conv_port_id(self, ce_id, cm_slot_num, ca_slot_num, ca_port_num):
        '''
        convert port info to standard string
        '''
        LOG.debug('_conv_port_id,Enter method')

        if ce_id is None:
            port_id = '%x%x%x' % (cm_slot_num, ca_slot_num, ca_port_num)
        else:
            port_id = '%x%x%x%x' % (ce_id,
                                    cm_slot_num,
                                    ca_slot_num,
                                    ca_port_num)
        # end of if

        LOG.debug('_conv_port_id,Exit method')
        return port_id

    def _wait_for_job_complete(self, conn, job):
        '''
        Given the job wait for it to complete.
        '''
        LOG.debug('_wait_for_job_complete,Enter method')

        self.retries = 0
        self.wait_for_job_called = False

        def _wait_for_job_complete():
            '''Called at an interval until the job is finished'''
            if self._is_job_finished(conn, job):
                raise loopingcall.LoopingCallDone()
            if self.retries > JOB_RETRIES:
                LOG.error(_LE("_wait_for_job_complete failed after %(retries)d"
                              " tries") % {'retries': self.retries})
                raise loopingcall.LoopingCallDone()
            # end of if
            try:
                self.retries += 1
                if not self.wait_for_job_called:
                    if self._is_job_finished(conn, job):
                        self.wait_for_job_called = True
            except Exception as e:
                LOG.error(_LE("Exception: %s") % six.text_type(e))
                exceptionMessage = (_LE("Issue encountered waiting for job."))
                LOG.error(exceptionMessage)
                raise exception.VolumeBackendAPIException(exceptionMessage)
        # end of def

        self.wait_for_job_called = False
        timer = loopingcall.FixedIntervalLoopingCall(_wait_for_job_complete)
        timer.start(interval=JOB_INTERVAL_SEC).wait()

        jobInstanceName = job['Job']
        jobinstance = conn.GetInstance(jobInstanceName,
                                       LocalOnly=False)

        rc = jobinstance['ErrorCode']

        LOG.debug('_wait_for_job_complete(%s),Exit method' % str(rc))
        return rc

    def _is_job_finished(self, conn, job):
        '''
        Check if the job is finished.
        '''
        LOG.debug('_is_job_finished,Enter method')

        jobInstanceName = job['Job']
        jobinstance = conn.GetInstance(jobInstanceName,
                                       LocalOnly=False)
        jobstate = jobinstance['JobState']
        LOG.debug('_is_job_finished,'
                  'state: %(state)s'
                  % {'state': jobstate})
        # From ValueMap of JobState in CIM_ConcreteJob
        # 2L=New, 3L=Starting, 4L=Running, 32767L=Queue Pending
        # ValueMap("2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13..32767,
        # 32768..65535"),
        # Values("New, Starting, Running, Suspended, Shutting Down,
        # Completed, Terminated, Killed, Exception, Service,
        # Query Pending, DMTF Reserved, Vendor Reserved")]
        # NOTE(deva): string matching based on
        #             http://ipmitool.cvs.sourceforge.net/
        #               viewvc/ipmitool/ipmitool/lib/ipmi_chassis.c

        if jobstate in [2L, 3L, 4L]:
            job_finished = False
        else:
            job_finished = True
        # end of if
        LOG.debug('_is_job_finished (finish:%s),'
                  'Exit method' % str(job_finished))
        return job_finished

    @eternus_dx_lockutils.FJDXLockutils('consume_pool', 'cinder-', True)
    def _get_pool_from_scheduler(self, volume_size_gb, use_tpp=True):
        LOG.debug('_get_pool_from_scheduler,Enter method')

        self.update_volume_stats()
        pool_instance = self.scheduler.consume_poolspace(
            volume_size_gb, use_tpp)

        LOG.debug('_get_pool_from_scheduler,Exit method')
        return pool_instance
