# Copyright (c) 2015 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

'''
FibreChannel Cinder Volume driver for Fujitsu ETERNUS DX S3 series.
'''

from oslo_log import log as logging
from cinder.volume import driver
from cinder.volume.drivers.fujitsu import eternus_dx_common
import six
from cinder.i18n import _LI

LOG = logging.getLogger(__name__)


class FJDXFCDriver(driver.FibreChannelDriver):
    '''
    ETERNUS FC Cinder Volume Driver
    for Fujitsu ETERNUS DX S3 series
    '''

    def __init__(self, *args, **kwargs):

        super(FJDXFCDriver, self).__init__(*args, **kwargs)
        self.common = eternus_dx_common.FJDXCommon(
            'fc',
            configuration=self.configuration)
        self.VERSION = self.common.VERSION
        return

    def check_for_setup_error(self):
        return

    def create_volume(self, volume):
        '''
        Create volume
        '''
        LOG.info(_LI('create_volume,Enter method'))

        (location, metadata) = self.common.create_volume(volume)

        v_metadata = volume.get('volume_metadata')
        for data in v_metadata:
            metadata[data['key']] = data['value']
        # end of for v_metadata

        LOG.info(_LI('create_volume,Exit method'))

        return {'provider_location': six.text_type(location),
                'metadata': metadata}

    def create_volume_from_snapshot(self, volume, snapshot):
        '''
        Creates a volume from a snapshot.
        '''
        LOG.info(_LI('create_volume_from_snapshot,Enter method'))

        (location, metadata) = \
            self.common.create_volume_from_snapshot(volume, snapshot)

        v_metadata = volume.get('volume_metadata')
        for data in v_metadata:
            metadata[data['key']] = data['value']
        # end of for v_metadata

        LOG.info(_LI('create_volume_from_snapshot,Exit method'))

        return {'provider_location': six.text_type(location),
                'metadata': metadata}

    def create_cloned_volume(self, volume, src_vref):
        '''
        Create cloned volume.
        '''
        LOG.info(_LI('create_cloned_volume,Enter method'))

        (location, metadata) = \
            self.common.create_cloned_volume(volume, src_vref)

        v_metadata = volume.get('volume_metadata')
        for data in v_metadata:
            metadata[data['key']] = data['value']
        # end of for v_metadata

        LOG.info(_LI('create_cloned_volume,Exit method'))

        return {'provider_location': six.text_type(location),
                'metadata': metadata}

    def delete_volume(self, volume):
        '''
        Delete volume
        '''
        LOG.info(_LI('delete_volume,Enter method'))

        self.common.delete_volume(volume)

        LOG.info(_LI('delete_volume,Exit method'))
        return

    def create_snapshot(self, snapshot):
        '''
        Creates a snapshot.
        '''
        LOG.info(_LI('create_snapshot,Enter method'))

        location = self.common.create_snapshot(snapshot)

        LOG.info(_LI('create_snapshot,Exit method'))
        return {'provider_location': six.text_type(location)}

    def delete_snapshot(self, snapshot):
        '''
        Deletes a snapshot.
        '''
        LOG.info(_LI('delete_snapshot,Enter method'))

        self.common.delete_snapshot(snapshot)

        LOG.info(_LI('delete_snapshot,Exit method'))
        return

    def ensure_export(self, context, volume):
        """Synchronously recreates an export for a volume."""
        return

    def create_export(self, context, volume, connector):
        """Exports the volume."""
        return

    def remove_export(self, context, volume):
        """Removes an export for a volume."""
        return

    def initialize_connection(self, volume, connector):
        '''
        Allow connection to connector and return connection info.
        '''
        LOG.info(_LI('initialize_connection,'
                     'volume id: %(volid)s,'
                     'connector info: %(wwpns)s,Enter method')
                 % {'volid': volume['id'],
                    'wwpns': connector['wwpns']})

        info = self.common.initialize_connection(volume, connector)

        LOG.info(_LI('initialize_connection,'
                     'connection info: %(info)s'
                     'Exit method') % {'info': info})
        return info

    def terminate_connection(self, volume, connector, **kwargs):
        '''
        Disallow connection from connector.
        '''
        LOG.info(_LI('terminate_connection,'
                     'volume id: %(volid)s,'
                     'connector info: %(wwpns)s,Enter method')
                 % {'volid': volume['id'],
                    'wwpns': connector['wwpns']})

        self.common.terminate_connection(volume, connector)

        LOG.info(_LI('terminate_connection,Exit method'))
        return

    def get_volume_stats(self, refresh=False):
        '''
        Get volume stats.
        If 'refresh' is True, run update the stats first.
        '''
        LOG.debug('get_volume_stats,Enter method')

        if refresh is True:
            data = self.common.update_volume_stats()
            backend_name = self.configuration.safe_get('volume_backend_name')
            data['volume_backend_name'] = backend_name or 'FJDXFCDriver'
            data['storage_protocol'] = 'fibre_channel'
            self._stats = data
        # end of if

        LOG.debug('get_volume_stats,Exit method')
        return self._stats

    def extend_volume(self, volume, new_size):
        '''
        Extend volume.
        '''
        LOG.info(_LI('extend_volume,Enter method'))

        self.common.extend_volume(volume, new_size)

        LOG.info(_LI('extend_volume,Exit method'))
        return
