# Copyright (c) 2015 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


from oslo_config import cfg
from cinder import exception
from oslo_utils import units
from oslo_log import log as logging
from cinder.i18n import _LE

LOG = logging.getLogger(__name__)

STD_VOLUME_COUNT_MAX = 128
POOL_TYPE_list = ['RAID', 'TPP']
FJ_ETERNUS_DX_OPT_list = [cfg.StrOpt('fujitsu_scheduler',
                                     default='default',
                                     help='scheduling system for '
                                     'selecting pool to create volume'),
                          cfg.BoolOpt('thin_provisioning_support',
                                      default=False,
                                      help='When True, return infinite value'
                                      ' for the free capacity of TPP')]


class FJDXScheduler(object):
    '''
    Pool Scheduler for ETERNUS OpenStack Cinder Volume Driver
    '''

    def __init__(self, configuration):
        scheduler_dic = {'default': self._skd_default,
                         'round-robin': self._skd_round_robin,
                         'capacity-first': self._skd_capacity_first}

        self.configuration = configuration
        self.configuration.append_config_values(FJ_ETERNUS_DX_OPT_list)
        self.pools = []
        self.pool_idx = -1
        self.pool_num = 0

        try:
            self.scheduler = \
                scheduler_dic.get(self.configuration.fujitsu_scheduler)
        except:
            msg = (_LE('Invalid scheduler was specified : %s')
                   % (self.configuration.fujitsu_scheduler))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        return

    def update_pools(self, pools):
        '''
        update pools
        '''
        LOG.debug('update_pools,Enter method')

        total_capacity_gb = 0
        free_capacity_gb = 0
        tpp_support = self.configuration.thin_provisioning_support

        self.pools = list(pools)
        self.pool_num = len(pools)

        if self.pool_num < self.pool_idx:
            self.pool_idx = 0
        # end of if

        for pool in pools:
            if pool['type'] == 'TPP':
                if tpp_support is True:
                    total_capacity_gb = 'infinite'
                    free_capacity_gb = 'infinite'
                    break
                # end of if
            # end of if
            total_capacity_gb += pool['total_capacity_gb']
            free_capacity_gb += pool['free_capacity_gb']
        # end of for pools

        # provide more informations
        pools_info = {'total_capacity_gb': total_capacity_gb,
                      'free_capacity_gb': free_capacity_gb}

        LOG.debug('update_pools, '
                  'pools: %(pools)s, '
                  'pools info: %(info)s,Exit method'
                  % {'pools': str(pools),
                     'info': pools_info})
        return pools_info

    def create_pool_info(self, pool_instance, volume_count, pool_type):
        '''
        create pool information from pool instance
        '''
        LOG.debug('create_pool_info,Enter method')

        if pool_type not in POOL_TYPE_list:
            msg = (_LE('Invalid pool type was specified : %s') % (pool_type))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        total_gb = pool_instance['TotalManagedSpace'] / units.Gi
        free_gb = pool_instance['RemainingManagedSpace'] / units.Gi
        pool = {'name': pool_instance['ElementName'],
                'path': pool_instance.path,
                'total_capacity_gb': total_gb,
                'free_capacity_gb': free_gb,
                'type': pool_type,
                'volume_count': volume_count}

        LOG.debug('create_pool_info,Exit method')
        return pool

    def consume_poolspace(self, volume_size_gb, use_tpp=True):
        '''
        consume pool space for volume in storage pool
        '''
        LOG.debug('consume_poolspace,Enter method')

        pools = []
        valid = False

        tpp_support = self.configuration.thin_provisioning_support

        # pools.append(None) to keep the pool size and
        # index the same as self.pools
        for pool in self.pools:
            if pool['type'] == 'RAID':
                if volume_size_gb < pool['free_capacity_gb']:
                    if pool['volume_count'] < STD_VOLUME_COUNT_MAX:
                        pools.append(pool)
                        valid = True
                    else:
                        pools.append(None)
                else:
                    pools.append(None)
                # end of if
            else:  # pool['type'] == 'TPP'
                if use_tpp is False:
                    pools.append(None)
                    continue
                if tpp_support is True:
                    pools.append(pool)
                    valid = True
                else:
                    if volume_size_gb < pool['free_capacity_gb']:
                        pools.append(pool)
                        valid = True
                    else:
                        pools.append(None)
                    # end of if
                # end of if
            # end of if
        # end of for self.pools

        if valid is False:
            msg = (_LE('No free space in all storage pool for created volume,'
                       'necessary space : %(volume_size)s')
                   % {'volume_size': volume_size_gb})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        # the size and index of pools is the same as self.pools
        self.pool_idx = self.scheduler(pools)
        selected_pool = self.pools[self.pool_idx]

        LOG.debug('consume_poolspace : '
                  'select pool(selected_pool:%(pool)s),Exit method'
                  % {'pool': str(selected_pool)})
        return selected_pool

    def update_poolspace(self, volume_size_gb, poolpath, mode):
        '''
        update pool space for volume in storage pool
        '''
        LOG.debug('update_poolspace,Enter method')

        for pool in self.pools:
            if pool['path']['InstanceID'] == poolpath['InstanceID']:
                selected_pool = pool
                break
            # end of if
        else:
            msg = (_LE('Specified pool is not exist : %s') % poolpath)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of for self.pools

        if mode is 'consume':  # create operation
            selected_pool['free_capacity_gb'] -= volume_size_gb
        elif mode is 'release':  # delete operation
            selected_pool['free_capacity_gb'] += volume_size_gb
        else:
            msg = (_LE('Specified mode is invalid : %s') % mode)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('update_poolspace,Exit method')
        return

    def _skd_default(self, pools):
        '''
        Scheduling System Functions
          get the index of the first pool which is not none
        '''

        idx = 0
        no = -1

        for no, pool in enumerate(pools):
            if pool is not None:
                idx = no
                break
            # end of if
        # end of for

        return idx

    def _skd_round_robin(self, pools):
        '''
        Scheduling System Functions
          begin at the index which is chose the last time
          get the index of the first pool which is not none
        '''

        idx = 0
        offset = self.pool_idx + 1

        for idx, pool in enumerate(pools[offset:]):
            if pool is not None:
                idx += offset
                break
            # end if
        else:
            for idx, pool in enumerate(pools[:offset]):
                if pool is not None:
                    break
                # end of if
            # end of for pools
        # end of for pools

        return idx

    def _skd_capacity_first(self, pools):
        '''
        Scheduling System Functions
          get the index of the max capacity pool
        '''

        idx = 0
        max_capacity = 0
        no = -1

        for no, pool in enumerate(pools):
            if pool is not None:
                capacity = pool['free_capacity_gb']
                if max_capacity < capacity:
                    idx = no
                    max_capacity = capacity
                # end of if
            # end of if
        # end of for pools

        return idx
