# Copyright (c) 2015 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

'''
Cinder Volume driver for Fujitsu ETERNUS DX S3 series.
This driver is extension version.
'''

# $Date:: 2015-09-17 22:12:32 +0900#$
# $Revision: 196 $

import os
import time
import threading
import uuid
import codecs
import six
from oslo_config import cfg
from xml.dom.minidom import parse as xml_parse
from xml.dom.minidom import Document
from xml.etree.ElementTree import parse as tree_parse
from cinder import context
from cinder import db
from cinder import exception
from cinder import utils
from oslo_log import log as logging
from oslo_concurrency import lockutils
from cinder.volume.drivers.fujitsu import eternus_dx_lockutils
from cinder.volume.drivers.fujitsu import eternus_dx_common
from cinder.i18n import _LE, _LI, _LW

LOG = logging.getLogger(__name__)

try:
    import pywbem
except:
    msg = (_LE('import pywbem failed!!'
               'pywbem is necessary for this volume driver.'))
    LOG.error(msg)
    raise exception.VolumeBackendAPIException(data=msg)

CONF = cfg.CONF
MONITOR_IMGVOL_INTERVAL = 600
DELETE_IMGVOL = "Deleting"
FJ_REMOTE_SRC_META = "FJ_Remote_Copy_Source"
FJ_REMOTE_DEST_META = "FJ_Remote_Copy_Destination"
FJ_REC_CLONE = "clone"
FJ_REC_MIRROR = "mirror"
FJ_VOL_FORMAT_KEY = "type:delete_with_volume_format"
LIMIT_SESSION_PER_VOLUME = 8  # limitation of OPC session by volume
FJ_SRC_VOL_ID = "FJ_Source_Volume_ID"
EC_REC = 3
RAIDGROUP = 2
TPPOOL = 5
CTRL_CONF = "FUJITSU_ControllerConfigurationService"

FJ_ETERNUS_DX_OPT_list = [cfg.StrOpt('fujitsu_min_image_volume_per_storage',
                                     default='0',
                                     help='minimum number of image volume '
                                     'per storage(ETERNUS)')]

CINDER_CONF_OPT_list = [cfg.StrOpt('fujitsu_image_management_dir',
                                   default=CONF.image_conversion_dir,
                                   help='directory for '
                                   'image management file(xml)')]
CONF.register_opts(CINDER_CONF_OPT_list)
FJ_QOS_KEY_list = ['maxBWS']

SPECIAL_STATS_dic = {'QoS_support': True}
POOL_TYPE_dic = {RAIDGROUP: 'RAID_GROUP', TPPOOL: 'Thinporvisioning_POOL'}
RETCODE_dic = {'32812': 'This operation cannot be performed to the Storage'
               ' Cluster resources'}


class FJDXCommonExt(eternus_dx_common.FJDXCommon):
    '''
    Common code that does not depend on protocol.
    '''

    def __init__(self, prtcl, configuration=None):

        self.supercls = super(FJDXCommonExt, self)
        self.supercls.__init__(prtcl, configuration)
        self.stats.update(SPECIAL_STATS_dic)
        self.configuration.append_config_values(FJ_ETERNUS_DX_OPT_list)
        self.configuration.fujitsu_image_management_file = \
            CONF.fujitsu_image_management_dir + \
            '/' + 'cinder_fujitsu_image_management.xml'
        imag_vol_num = self.configuration.fujitsu_min_image_volume_per_storage
        use_fj_imag_vol = self.configuration.use_fujitsu_image_volume
        if ((imag_vol_num.isdigit() is False) or (use_fj_imag_vol is False)):
            self.configuration.fujitsu_min_image_volume_per_storage = '0'
        # end of if

        self._check_user()
        self.invalid_migration_list = []
        return

    def create_volume(self, volume):
        '''
        create volume on ETERNUS
        '''
        LOG.debug('create_volume,Enter method')

        (element_path, metadata) = self.supercls.create_volume(volume)

        # set qos to created volume
        self._set_qos(volume)

        LOG.debug('create_volume,Exit method')
        return (element_path, metadata)

    def create_volume_from_snapshot(self, volume, snapshot):
        '''
        Creates a volume from a snapshot
        '''
        LOG.debug('create_volume_from_snapshot,Enter method')

        self.conn = self._get_eternus_connection()
        source_volume_instance = self._find_lun(snapshot)

        # Existence check the source volume
        if source_volume_instance is None:
            msg = (_LE('create_volume_from_snapshot,'
                       'Source Volume is not exist in ETERNUS.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        # create volume for destination of cloned volume
        (element_path, metadata) = self.create_volume(volume)
        target_volume_instancename = \
            self._create_eternus_instance_name(element_path['classname'],
                                               element_path['keybindings'])
        target_volume_instance = \
            self._get_eternus_instance(target_volume_instancename)
        self._create_local_cloned_volume(volume, target_volume_instance,
                                         snapshot, source_volume_instance)

        LOG.debug('create_volume_from_snapshot,Exit method')
        return (element_path, metadata)

    def create_cloned_volume(self, volume, src_vref,
                             CloneOnly=False, use_service_name=False):
        '''
        Create clone of the specified volume.
        '''
        LOG.debug('create_cloned_volume,Enter method')

        source_volume_instance = self._find_lun(src_vref, use_service_name)

        if source_volume_instance is None:
            msg = (_LE('create_cloned_volume,'
                       'Source Volume is not exist in ETERNUS.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        if CloneOnly is False:
            (element_path, metadata) = self.create_volume(volume)
            target_volume_instancename = \
                self._create_eternus_instance_name(element_path['classname'],
                                                   element_path['keybindings'])
            target_volume_instance = \
                self._get_eternus_instance(target_volume_instancename)
        else:
            target_volume_instance = self._find_lun(volume)
            try:
                element_path = eval(volume['provider_location'])
                metadata = volume['volume_metadata']
            except:
                element_path = None
                metadata = None
        # end of if

        remote_copy_type = \
            self._get_metadata(volume).get(FJ_REMOTE_DEST_META, None)
        if remote_copy_type is None:
            self._create_local_cloned_volume(volume,
                                             target_volume_instance,
                                             src_vref,
                                             source_volume_instance)
        else:
            source_volume_metadata = self._get_metadata(src_vref)
            self._create_remote_cloned_volume(
                volume, metadata, src_vref,
                source_volume_metadata, remote_copy_type)
        # end of if

        LOG.debug('create_cloned_volume,Exit method')

        return (element_path, metadata)

    def _create_remote_cloned_volume(self, volume, volume_metadata, src_vref,
                                     src_vref_metadata, remote_copy_type):
        '''
        Create remote clone of the specified volume.
        '''
        LOG.debug('_create_remote_cloned_volume,Enter method')

        ctxt = context.get_admin_context()
        copy_type = str(remote_copy_type).lower()

        # judge if source volume has already get a mirror target
        source_volume_copy_dict = \
            eval(src_vref_metadata.get(FJ_REMOTE_SRC_META, "{}"))

        # when remote_copy_type is mirror, check mirror copy already exist
        if copy_type == FJ_REC_MIRROR:
            for value in source_volume_copy_dict.values():
                if value == FJ_REC_MIRROR:
                    s_volumename = self._create_volume_name(src_vref['id'])
                    msg = (_LE('_create_remote_cloned_volume,'
                               'source volume:%(src)s already has a mirror!')
                           % {'src': s_volumename})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                # end of if
            # end of for source_volume_copy_dict.values()
        # end of if

        # execute ccm script
        self._exec_ccm_script("start",
                              source=src_vref_metadata,
                              source_volume=src_vref,
                              target=volume_metadata,
                              copy_type=copy_type)

        # update metadata of source volume
        lockname = 'db-update-' + src_vref['id']

        @eternus_dx_lockutils.FJDXLockutils(lockname, 'cinder-', True)
        def __update_metadata(self):
            '''
            update volume metadata
            '''
            source_volume_copy_dict = {volume['id']: copy_type}
            metadata = \
                {FJ_REMOTE_SRC_META: six.text_type(source_volume_copy_dict)}
            db.volume_metadata_update(ctxt.elevated(), src_vref['id'],
                                      metadata, False)
        # end of def

        if copy_type == FJ_REC_MIRROR:
            # if remote copy type is mirror, update metadata
            __update_metadata(self)
        elif copy_type == FJ_REC_CLONE:
            # if remote copy type and copy_list is {}, update metadata
            if len(source_volume_copy_dict) == 0:
                __update_metadata(self)
            # end of if
        # end of if

        LOG.debug('_create_remote_cloned_volume,Exit method')
        return

    def delete_volume(self, volume):
        '''
        Delete volume on ETERNUS.
        '''
        LOG.debug('delete_volume,Enter method')

        recheck_session_flag = False

        vol_instance = self._find_lun(volume)
        if vol_instance is None:
            return
        # end of if

        # CCM processing
        metadata = self._get_metadata(volume)
        copy_type = metadata.get(FJ_REMOTE_DEST_META, None)

        if (copy_type is not None) and ("FJ_Backend" in metadata):
            self._exec_ccm_script("stop", target=metadata)
            recheck_session_flag = True

            # if copy_type is mirror update FJ_Remote_Copy_Source info
            tmp_flag1 = FJ_SRC_VOL_ID in metadata
            tmp_flag2 = str(copy_type).lower() == FJ_REC_MIRROR
            if tmp_flag1 and tmp_flag2:
                source_volume_id = metadata.get(FJ_SRC_VOL_ID)
                ctxt = context.get_admin_context()
                src_vref = db.volume_get(ctxt, source_volume_id)
                source_volume_metadata = self._get_metadata(src_vref)
                source_volume_copy_list = \
                    eval(source_volume_metadata.get(FJ_REMOTE_SRC_META, "{}"))

                # if source_volume_copy_list has this volume delete it
                volume_id = volume['id']
                if volume_id in source_volume_copy_list:
                    del source_volume_copy_list[volume_id]
                # end of if

                update_metadata = \
                    {FJ_REMOTE_SRC_META:
                     six.text_type(source_volume_copy_list)}
                db.volume_metadata_update(ctxt.elevated(),
                                          source_volume_id,
                                          update_metadata,
                                          False)
            # end of if
        # end of if

        if metadata.get(FJ_REMOTE_SRC_META, None) is not None:
            self._exec_ccm_script(
                "stop", source=metadata, source_volume=volume)
            recheck_session_flag = True

            # update volume metadata
            copy_list = eval(metadata.get(FJ_REMOTE_SRC_META))
            ctxt = context.get_admin_context()
            for target_volume_id, copy_type in copy_list.items():
                try:
                    if copy_type == FJ_REC_MIRROR:
                        db.volume_metadata_update(
                            ctxt.elevated(), target_volume_id,
                            {FJ_REMOTE_DEST_META: None}, False)
                    # end of if
                except:
                    pass
            # end of for copy_list
        # end of if

        valid = self._delete_volume_setting(
            volume, recheck_flag=recheck_session_flag)

        if valid is False:
            return
        # end of if

        vol_instance = self._find_lun(volume)
        if vol_instance is None:
            return
        # end of if

        with_format = self._get_extra_specs(volume,
                                            key=FJ_VOL_FORMAT_KEY,
                                            default='False')
        if self._get_bool(with_format) is True:
            pool_name, pool = self._find_pool_from_volume(volume, vol_instance)
            if pool is None:
                msg = (_LE('delete_volume,'
                           'eternus_pool:%(eternus_pool)s,'
                           'not found.')
                       % {'eternus_pool': pool_name})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            if 'RSP' in pool['InstanceID']:  # pooltype == RAIDGROUP
                self._format_standard_volume(volume)
            else:  # pooltype == TPPOOL
                self._format_tpv(volume, vol_instance)
            # end of if
        # end of if

        self._delete_volume(vol_instance)

        LOG.debug('delete_volume,Exit method')
        return

    def _delete_volume_setting(self, volume, recheck_flag=False):
        '''
        Delete volume setting ( HostAffinity, CopySession) on ETERNUS.
        '''
        LOG.debug('_delete_volume_setting,Enter method')

        # recheck ccm copy session status
        if recheck_flag is True:

            # Existence check the volume
            volumename = self._create_volume_name(volume['id'])
            vol_instance = self._find_lun(volume)
            if vol_instance is None:
                LOG.info(_LI('_delete_volume_setting,'
                             'volumename:%(volumename)s,'
                             'volume not found on ETERNUS.'
                             'delete only management data on cinder database.')
                         % {'volumename': volumename})
                return False
            # end of if

            cpsession_instance = None
            cpsessionlist = self._find_copysession(vol_instance)
            for cpsession in cpsessionlist:
                while True:
                    try:
                        # get copy session instance
                        cpsession_instance = \
                            self.conn.GetInstance(cpsession, LocalOnly=False)
                    except Exception:
                        cpsession_instance = None

                    # if None break(session not exist)
                    if cpsession_instance is None:
                        break
                    # end of if

                    # if CopyType is not ccm operation type break(EC/REC=3)
                    if cpsession_instance['CopyType'] != EC_REC:
                        break
                    # end of if

                    # CopyState 2-Initialized/3-Unsynchronized(need to wait)
                    if cpsession_instance['CopyState'] not in [2, 3]:
                        break
                    # end of if

                    time.sleep(5)
                # end of while
            # end of for cpsessionlist
        # end of if

        valid = self.supercls._delete_volume_setting(volume)

        LOG.debug('_delete_volume_setting,Exit method')
        return valid

    def _format_standard_volume(self, volume):
        '''
        Format standard volume
        '''
        LOG.debug('_format_standard_volume,Enter method')

        wait_sec = 60  # 1min
        wait_cnt = 0

        volumename = self._create_volume_name(volume['id'])
        param_dict = {'volume-name': volumename}

        rc, errordesc, job = \
            self._exec_eternus_cli('format_volume',
                                   **param_dict)

        if rc != 0L:
            msg = (_LE('_format_standard_volume,'
                       'volumename:%(volumename)s,'
                       'Return code:%(rc)lu,'
                       'Error:%(errordesc)s,'
                       'Job:%(job)s')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc,
                      'job': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        while True:
            time.sleep(wait_sec)
            rc, errordesc, job = \
                self._exec_eternus_cli("show_volume_progress", **param_dict)

            if rc != 0L:
                msg = (_LE('_format_standard_volume,'
                           'show volume progress error,'
                           'volumename:%(volumename)s,'
                           'Return code:%(rc)lu,'
                           'Error:%(errordesc)s,'
                           'Job:%(job)s')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'job': job})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            wait_sec = int(job)

            if wait_sec == 0:
                LOG.debug('_format_standard_volume,'
                          'format complete')
                break
            elif wait_cnt < 3:
                wait_sec /= 3
                wait_cnt += 1
            # end of if

            LOG.debug('_format_standard_volume,'
                      'format remain time : %s sec' % job)

        LOG.debug('_format_standard_volume,Exit method')
        return

    def _format_tpv(self, volume, vol_instance, poolname=None):
        '''
        Format tpv
        '''
        LOG.debug('_format_tpv,Enter method')

        volumename = self._create_volume_name(volume['id'])

        if poolname is None:
            v_metadata = volume.get('volume_metadata')
            for data in v_metadata:
                if data['key'] == "FJ_Pool_Name":
                    poolname = data['value']
                    break
        # end of for v_metadata
        # end of if

        lockname = 'format-' + poolname

        @eternus_dx_lockutils.FJDXLockutils(lockname, 'cinder-', True)
        def __format_tpv(self, volumename, poolname):
            '''
            do format tpv using CLI method
            '''

            wait_sec = 60  # 1min
            wait_cnt = 0

            format_param_dict = {'volume-name': volumename}

            rc, errordesc, job = \
                self._exec_eternus_cli('format_tpv',
                                       **format_param_dict)

            if rc != 0L:
                msg = (_LE('_format_tpv,'
                           'volumename:%(volumename)s,'
                           'Return code:%(rc)lu,'
                           'Error:%(errordesc)s,'
                           'Job:%(job)s')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'job': job})

                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            show_param_dict = {'pool-name': poolname}
            while True:
                time.sleep(wait_sec)
                rc, errordesc, job = \
                    self._exec_eternus_cli("show_tpv_progress",
                                           **show_param_dict)

                if rc != 0L:
                    msg = (_LE('_format_tpv,'
                               'show volume progress error,'
                               'volumename:%(volumename)s,'
                               'Return code:%(rc)lu,'
                               'Error:%(errordesc)s,'
                               'Job:%(job)s')
                           % {'volumename': volumename,
                              'rc': rc,
                              'errordesc': errordesc,
                              'job': job})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                # end of if

                wait_sec = int(job)

                if wait_sec == 0:
                    LOG.debug('_format_tpv,format complete')
                    break
                elif wait_cnt < 3:
                    wait_sec /= 3
                    wait_cnt += 1
                # end of if

                LOG.debug('_format_tpv,format wait : %s sec' % job)
                # end of if
            # end of while
        # end of def
        __format_tpv(self, volumename, poolname)

        LOG.debug('_format_tpv,Exit method')
        return

    def initialize_connection(self, volume, connector):
        '''
        Allow connection to connector and return connection info.
        '''
        LOG.debug('initialize_connection,Enter method')

        # CCM processing
        metadata = self._get_metadata(volume)
        if metadata.get(FJ_REMOTE_DEST_META, None) == FJ_REC_MIRROR:
            if volume['status'] == "in-use":
                msg = (_LE('Live migration for backup volume is not allowed'))
                self.invalid_migration_list.append(volume['id'])
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            self._exec_ccm_script("suspend", target=metadata)
        # end of if

        try:
            vol_instance = self._find_lun(volume)
            # Existence check the volume
            if vol_instance is None:
                # volume does not found
                msg = (_LE('initialize_connection,'
                           'volume:%(volume)s,'
                           'Volume not found.')
                       % {'volume': volume['name']})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            self.conn = self._get_eternus_connection()
            target_portlist, portidlist = self._get_target_port_info()
            mapdata = self._get_mapdata(vol_instance,
                                        connector, target_portlist)

            if mapdata is not None:
                # volume is already mapped
                target_lun = None
                if 'target_lun' in mapdata:
                    target_lun = mapdata['target_lun']
                else:
                    target_lun = mapdata['target_luns'][0]
                # target_lun = mapdata['target_lun'] if 'target_lun' in mapdata
                # else mapdata['target_luns'][0]
                msg = (_LI('initialize_connection,'
                           'volume:%(volume)s,'
                           'target_lun:%(target_lun)s,'
                           'Volume is already mapped.')
                       % {'volume': volume['name'],
                          'target_lun': target_lun})
                LOG.info(msg)
            else:
                self._map_lun(vol_instance, connector, portidlist)
                mapdata = self._get_mapdata(vol_instance,
                                            connector,
                                            target_portlist)
            # end of if

            mapdata['target_discoverd'] = True
            mapdata['volume_id'] = volume['id']

            if self.protocol == 'fc':
                device_info = {'driver_volume_type': 'fibre_channel',
                               'data': mapdata}
            elif self.protocol == 'iSCSI':
                device_info = {'driver_volume_type': 'iscsi',
                               'data': mapdata}
            # end of if

            LOG.debug('initialize_connection,'
                      'device_info:%(info)s,'
                      'Exit method'
                      % {'info': device_info})
        except Exception as ex:
            # when volume is set to REC Mirror, resume the session
            if metadata.get(FJ_REMOTE_DEST_META, None) == FJ_REC_MIRROR:
                self._exec_ccm_script("resume", target=metadata)
                raise ex
            # end of if

        return device_info

    def terminate_connection(self, volume, connector,
                             force=False, **kwargs):
        '''
        Disallow connection from connector
        '''
        LOG.debug('terminate_connection,Enter method')

        if volume['id'] in self.invalid_migration_list:
            msg = (_LI('Live migration for backup volume is not allowed, '
                       'Terminate Connection Skip'))
            self.invalid_migration_list.remove(volume['id'])
            LOG.info(msg)
            return
        # end of if

        self.supercls.terminate_connection(volume, connector)

        # CCM processing
        metadata = self._get_metadata(volume)
        if metadata.get(FJ_REMOTE_DEST_META, None) == FJ_REC_MIRROR:
            self._exec_ccm_script("resume", target=metadata)
        # end of if

        LOG.debug('terminate_connection,Exit method')
        return

    def extend_volume(self, volume, new_size):
        '''
        extend volume on ETERNUS
        '''
        LOG.debug('extend_volume,Enter method')

        # create to volumename on ETERNUS from cinder VolumeID
        volumename = self._create_volume_name(volume['id'])

        # get source volume instance
        source_volume_instance = self._find_lun(volume)
        if source_volume_instance is None:
            msg = (_LE('extend_volume,'
                       'source_volume:%(volumename)s,'
                       'not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        pathStr = str(source_volume_instance.path)
        LOG.debug('extend_volume,volumename:%(volumename)s,'
                  'volumesize:%(volumesize)u,'
                  'source volume instance:%(source_volume_instance)s'
                  % {'volumename': volumename,
                     'volumesize': new_size,
                     'source_volume_instance': pathStr})

        self.conn = self._get_eternus_connection()

        # get poolname from driver configuration file
        pool_name, pool = self._find_pool_from_volume(volume)

        # Existence check the pool
        if pool is None:
            msg = (_LE('extend_volume,'
                       'eternus_pool:%(eternus_pool)s,'
                       'not found.')
                   % {'eternus_pool': pool_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        # set pooltype
        if 'RSP' in pool['InstanceID']:
            pooltype = RAIDGROUP
        else:
            pooltype = TPPOOL
        # end of if

        if pooltype == RAIDGROUP:
            extend_size = str(new_size - volume['size']) + 'gb'
            param_dict = {'volume-name': volumename,
                          'rg-name': pool_name,
                          'size': extend_size}
            rc, errordesc, job = self._exec_eternus_cli('expand_volume',
                                                        **param_dict)

            if rc != 0L:
                msg = (_LE('extend_volume,'
                           'volumename:%(volumename)s,'
                           'Return code:%(rc)lu,'
                           'Error:%(errordesc)s,'
                           'PoolType:%(pooltype)s')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'pooltype': POOL_TYPE_dic[pooltype]})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            LOG.debug('extend_volume,'
                      'volumename:%(volumename)s,'
                      'Return code:%(rc)lu,'
                      'Error:%(errordesc)s,'
                      'Pool Name:%(eternus_pool)s,'
                      'Pool Type:%(pooltype)s'
                      % {'volumename': volumename,
                         'rc': rc,
                         'errordesc': errordesc,
                         'eternus_pool': pool_name,
                         'pooltype': POOL_TYPE_dic[pooltype]})

        else:  # pooltype == TPPOOL
            self.supercls.extend_volume(volume, new_size)
        # end of if

        LOG.debug('extend_volume,Exit method')
        return

    def create_image_volume(self, volume_ref, image_id):
        '''
        create image volume
        '''
        LOG.debug('create_image_volume,Enter method')

        img_volume = {}

        img_volume['id'] = 'image-' + uuid.uuid4().hex
        img_volume['display_name'] = volume_ref['display_name']
        img_volume['size'] = volume_ref['size']
        img_volume['volume_type_id'] = volume_ref['volume_type_id']
        img_volume['volume_metadata'] = []
        storage_name = self._get_drvcfg('EternusIP')

        LOG.info(_LI('create_image_volume, '
                     'image volume id:%(img_volid)s, '
                     'storage name:%(storage_name)s')
                 % {'img_volid': img_volume['id'],
                    'storage_name': storage_name})

        (element_path, metadata) = self.create_volume(img_volume)
        img_volume['provider_location'] = six.text_type(element_path)

        LOG.debug('create_image_volume,Exit method')
        return img_volume

    @lockutils.synchronized('ETERNUS_DX-img-update', 'cinder-', True)
    def add_image_volume(self, volume_ref, image_id):
        '''
        add image volume
        '''
        LOG.debug('add_image_volume,Enter method')

        storage_name = self._get_drvcfg('EternusIP')

        with_format = self._get_extra_specs(volume_ref,
                                            key=FJ_VOL_FORMAT_KEY,
                                            default=False)
        use_format = self._get_bool(with_format)

        LOG.debug('add_image_volume,'
                  'image_id:%(image_id)s,'
                  'volume_id:%(volume_id)s,'
                  'size:%(size)s,'
                  'storage_name:%(storage_name)s'
                  % {'image_id': image_id,
                     'volume_id': volume_ref['id'],
                     'size': volume_ref['size'],
                     'storage_name': storage_name})

        self._add_image_volume_info(volume_ref['id'], volume_ref['size'],
                                    volume_ref['provider_location'],
                                    image_id, storage_name, use_format)

        LOG.debug('add_image_volume,Exit method')
        return

    def _delete_image_volume(self, volume, image_id,
                             format_volume=False, delete_volume=True):
        '''
        delete image volume
        '''
        LOG.debug('_delete_image_volume,Enter method')

        @lockutils.synchronized('ETERNUS_DX-img-update', 'cinder-', True)
        def delete_image_volume_prepare(self):
            '''
            set delete mark to image volume management file
            '''
            self._update_image_volume_info(volume['id'],
                                           image_id, value=DELETE_IMGVOL)
        # end of def

        @lockutils.synchronized('ETERNUS_DX-img-update', 'cinder-', True)
        def delete_image_volume_update(self):
            '''
            remove image volume info from image volume management file
            '''
            self._update_image_volume_info(volume['id'], image_id, remove=True)
        # end of def

        delete_image_volume_prepare(self)

        if delete_volume is True:
            if format_volume is True:
                vol_instance = self._find_lun(volume)

                try:
                    AssocClassStr = 'FUJITSU_AllocatedFromStoragePool'
                    pool = \
                        self._assoc_eternus(vol_instance.path,
                                            AssocClass=AssocClassStr,
                                            ResultClass='CIM_StoragePool')[0]
                except:
                    msg = (_LE('_delete_image_volume,'
                               'vol_instance.path:%(vol)s')
                           % {'vol': vol_instance.path})

                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                if 'RSP' in pool['InstanceID']:
                    # pooltype = RAIDGROUP
                    self._format_standard_volume(volume)
                else:
                    # pooltype = TPPOOL
                    self._format_tpv(volume, vol_instance, pool['ElementName'])
                # end of if
            # end of if

            self.delete_volume(volume)
        # end of if

        delete_image_volume_update(self)

        LOG.debug('_delete_image_volume,Exit method')
        return

    @lockutils.synchronized('ETERNUS_DX-img-update', 'cinder-', True)
    def copy_image_volume_to_volume(self, volume, image_id, src_vref=None):
        '''
        copy image volume to volume
        '''
        LOG.debug('copy_image_volume_to_volume,Enter method')

        cloned = False
        src_volume_id = None

        # get image volume path which meets new volume's condition
        if src_vref is None:
            image_management_file = \
                self.configuration.fujitsu_image_management_file
            storage_name = self._get_drvcfg('EternusIP')
            src_vref = self._get_imgcfg(image_management_file,
                                        volume['size'], image_id, storage_name)

            if src_vref is not None:
                src_volume_id = src_vref['id']

            LOG.debug('copy_image_volume_to_volume,'
                      'image_management_file_name:%(image_management_file)s,'
                      'src_volume_id:%(src_volume_id)s,'
                      'storage_name:%(storage_name)s'
                      % {'image_management_file': image_management_file,
                         'src_volume_id': src_volume_id,
                         'storage_name': storage_name})
        else:
            src_volume_id = src_vref['id']
            LOG.debug('copy_image_volume_to_volume,'
                      'src_volume_id:%(src_volume_id)s'
                      % {'src_volume_id': src_volume_id})
        # end of if

        # copy image volume to volume using above image volume path
        if src_volume_id is not None:
            self.create_cloned_volume(volume, src_vref, CloneOnly=True)
            self._update_image_volume_info(src_volume_id, image_id)
            cloned = True
        # end of if

        LOG.debug('copy_image_volume_to_volume '
                  '(cloned=%s),Exit method' % (cloned))
        return cloned

    def monitor_image_volume(self):
        '''
        timer function for monitor image volume
        '''
        try:
            self._monitor_image_volume()
        except Exception as e:
            LOG.warn(_LW('monitor_image_volume, '
                         'undefined error was occured (%s)') % str(e))

        timer = threading.Timer(MONITOR_IMGVOL_INTERVAL,
                                self.monitor_image_volume)
        timer.start()
        return

    @lockutils.synchronized('ETERNUS_DX-img-monitor', 'cinder-', True)
    def _monitor_image_volume(self):
        '''
        monitor image volume
        '''
        LOG.debug('monitor_image_volume, Enter method')

        format = False

        image_management_file = \
            self.configuration.fujitsu_image_management_file
        storage_name = self._get_drvcfg('EternusIP')
        nosession_volume_limit = \
            int(self.configuration.fujitsu_min_image_volume_per_storage)

        if os.path.exists(image_management_file):
            LOG.debug('monitor_image_volume,'
                      'image_management_file:%(image_management_file)s,'
                      'storage_name:%(storage_name)s,'
                      'nosession_volume_limit:%(nosession_volume_limit)s'
                      % {'image_management_file': image_management_file,
                         'storage_name': storage_name,
                         'nosession_volume_limit': nosession_volume_limit})
            doc = xml_parse(image_management_file)
            f_image = doc.getElementsByTagName('Image')

            for f_img in f_image:
                f_image_id = \
                    f_img.getElementsByTagName('ImageID')[0].childNodes[0].data
                f_volume = f_img.getElementsByTagName('Volume')
                nosession_volume = 0
                for f_vol in f_volume:
                    f_stor = f_vol.getElementsByTagName('StorageName')[0]
                    f_storage_name = f_stor.childNodes[0].data
                    if storage_name != f_storage_name:
                        continue
                    # end of if

                    f_vid = f_vol.getElementsByTagName('VolumeID')[0]
                    f_volume_id = f_vid.childNodes[0].data

                    try:
                        f_vpth = f_vol.getElementsByTagName('VolumePath')[0]
                        f_volume_path = f_vpth.childNodes[0].data
                    except:
                        f_volume_path = None

                    volume = {'id': f_volume_id,
                              'provider_location': f_volume_path}

                    try:
                        f_fmat = f_vol.getElementsByTagName('Format')[0]
                        f_format = f_fmat.childNodes[0].data
                        format = self._get_bool(f_format)
                    except:
                        pass

                    try:
                        session_num = self._get_sessionnum_by_srcvol(volume)
                        if session_num == 0:
                            nosession_volume += 1
                            if nosession_volume_limit < nosession_volume:
                                LOG.info(_LI('monitor_image_volume,'
                                             'delete unused image volume which'
                                             ' is not managed by cinder and is'
                                             ' made to boost performance,'
                                             'volume id:%(volumeid)s')
                                         % {'volumeid': f_volume_id})

                                self._delete_image_volume(volume,
                                                          f_image_id,
                                                          format_volume=format)
                                continue
                            # end of if
                        # end of if

                        self._update_image_volume_info(f_volume_id,
                                                       f_image_id,
                                                       str(session_num))
                    except Exception as e:
                        LOG.info(_LI('monitor_image_volume,'
                                     'image volume update '
                                     'event : %(id)s (%(err)s)')
                                 % {'id': f_volume_id, 'err': str(e)})

                        self._delete_image_volume(volume,
                                                  f_image_id,
                                                  delete_volume=False)
                # end of for f_volume
            # end of for f_image

        # end of if
        LOG.debug('monitor_image_volume, Exit method')
        return

    def _get_imgcfg(self, filename, volume_size, image_id, storage_name):
        '''
        read from image management file.
        '''
        LOG.debug("_get_imgcfg input[%s][%s][%s],Enter method"
                  % (filename, image_id, str(volume_size)))

        ret = None

        if os.path.exists(filename):
            try:
                tree = tree_parse(filename)
                elem = tree.getroot()

                for child in elem:
                    f_image_id = child.findtext(".//ImageID")

                    if image_id != f_image_id:
                        continue
                    # end of if

                    volumes = child.findall(".//Volume")
                    for volume in volumes:
                        f_session_text = volume.findtext(".//Session")
                        if f_session_text == DELETE_IMGVOL:
                            continue
                        # end of if

                        f_session_num = int(f_session_text)
                        f_volume_size = int(volume.findtext(".//VolumeSize"))
                        f_storage_name = volume.findtext(".//StorageName")

                        try:
                            f_volume_path = volume.findtext(".//VolumePath")
                        except:
                            f_volume_path = None

                        if (f_session_num < LIMIT_SESSION_PER_VOLUME) and \
                           (f_volume_size <= volume_size) and \
                           (storage_name == f_storage_name):
                            ret = {'id': volume.findtext(".//VolumeID"),
                                   'provider_location': f_volume_path}
                            break
                        else:
                            continue
                        # end of if
                    # end of for volumes

                    if ret is not None:
                        break
                    # end of if
                # end of for elem
            except:
                LOG.info(_LI("_get_imgcfg, management file is invalid."))
        # end of if

        LOG.debug("_get_imgcfg output[%s],Exit method" % (str(ret)))
        return ret

    def _exec_ccm_script(self, command, source=None, source_volume=None,
                         target=None, target_volume=None, copy_type=None):
        '''
        Execute ccm script
        '''
        LOG.debug('_exec_ccm_script,Enter method')

        s_option = ""
        d_option = ""
        t_option = ""

        if source is not None:
            if 'FJ_Volume_No' in source:
                s_olu_no = source['FJ_Volume_No']
            else:
                vol_instance = \
                    self._find_lun(source_volume, use_service_name=True)
                s_olu_no = "0x" + vol_instance['DeviceID'][24:28]
            # end of if

            s_serial_no = source['FJ_Backend'][-10:]
            s_option = " -s \"%s/%s\"" % (s_serial_no, s_olu_no)
            command += s_option
        # end of if

        if target is not None:
            d_olu_no = target['FJ_Volume_No']
            d_serial_no = target['FJ_Backend'][-10:]
            d_option = " -d \"%s/%s\"" % (d_serial_no, d_olu_no)
            command += d_option
        # end of if

        if copy_type is not None:
            t_option = " -t %s" % ((copy_type[0]).lower() + copy_type[1:])
            command += t_option
        # end of if

        try:
            (out, _err) = utils.execute('acrec', command, run_as_root=True)
            LOG.debug('_exec_ccm_script,'
                      'command:%(command)s,'
                      's_option:%(s_option)s,'
                      'd_option:%(d_option)s,'
                      't_option:%(t_option)s'
                      % {'command': command,
                         's_option': s_option,
                         'd_option': d_option,
                         't_option': t_option})

        except Exception as ex:
            LOG.error(_LE(ex))
            LOG.error(_LE('_exec_ccm_script,'
                          'command:%(command)s,'
                          's_option:%(s_option)s,'
                          'd_option:%(d_option)s,'
                          't_option:%(t_option)s')
                      % {'command': command, 's_option': s_option,
                         'd_option': d_option, 't_option': t_option})
            raise exception.VolumeBackendAPIException(ex)

        LOG.debug('_exec_ccm_script,Exit method')
        return

    def _add_image_volume_info(self, volume_id, volume_size, volume_path,
                               image_id, storage_name, use_format):
        '''
        add image volume information to image management file
        '''
        LOG.debug('_add_image_volume_info,Enter method')

        image_management_file = \
            self.configuration.fujitsu_image_management_file

        # if file is not exist, then make formatted file
        if not os.path.exists(image_management_file):
            LOG.debug('_add_image_volume_info, create new management file')
            doc = Document()
            root = doc.createElement('FUJITSU')
            doc.appendChild(root)

            f = codecs.open(image_management_file, 'wb', 'UTF-8')
            doc.writexml(f, '', '  ', '\n', encoding='UTF-8')
            f.close()
        # end of if

        # add image volume information
        doc = xml_parse(image_management_file)
        root = doc.documentElement
        image = doc.getElementsByTagName('Image')

        for img in image:
            f_image_id = \
                img.getElementsByTagName('ImageID')[0].childNodes[0].data
            if image_id != f_image_id:
                continue
            # end of if

            volume = doc.createElement('Volume')
            img.appendChild(doc.createTextNode('\n   '))
            img.appendChild(volume)
            break
        else:
            image = doc.createElement('Image')
            root.appendChild(doc.createTextNode('\n '))
            root.appendChild(image)

            f_image_id = doc.createElement('ImageID')
            image.appendChild(f_image_id)
            f_image_id.appendChild(doc.createTextNode(image_id))

            volume = doc.createElement('Volume')
            image.appendChild(doc.createTextNode('\n   '))
            image.appendChild(volume)
        # end of for image

        f_volume_id = doc.createElement('VolumeID')
        volume.appendChild(f_volume_id)
        f_volume_id.appendChild(doc.createTextNode(volume_id))

        f_volume_size = doc.createElement('VolumeSize')
        volume.appendChild(f_volume_size)
        f_volume_size.appendChild(doc.createTextNode(str(volume_size)))

        f_volume_path = doc.createElement('VolumePath')
        volume.appendChild(f_volume_path)
        f_volume_path.appendChild(doc.createTextNode(volume_path))

        f_storage_name = doc.createElement('StorageName')
        volume.appendChild(f_storage_name)
        f_storage_name.appendChild(doc.createTextNode(storage_name))

        f_session = doc.createElement('Session')
        volume.appendChild(f_session)
        f_session.appendChild(doc.createTextNode('0'))

        f_format = doc.createElement('Format')
        volume.appendChild(f_format)
        f_format.appendChild(doc.createTextNode(str(use_format)))

        f = codecs.open(image_management_file, 'wb', 'UTF-8')
        doc.writexml(f, encoding='UTF-8')
        f.close()
        LOG.debug('_add_image_volume_info'
                  'image_management_file:%(image_management_file)s,'
                  'image_id:%(image_id)s,'
                  'volume_id:%(volume_id)s,'
                  'volume_size:%(volume_size)s,'
                  'storage_name:%(storage_name)s'
                  % {'image_management_file': image_management_file,
                     'image_id': image_id,
                     'volume_id': volume_id,
                     'volume_size': str(volume_size),
                     'storage_name': storage_name})
        LOG.debug('_add_image_volume_info,Exit method')

    def _update_image_volume_info(self,
                                  volume_id,
                                  image_id, value=None, remove=False):
        '''
        update image volume information in image management file
        '''
        LOG.debug('_update_image_volume_info,Enter method')

        f_image = None
        f_volume = None

        image_management_file = \
            self.configuration.fujitsu_image_management_file
        doc = xml_parse(image_management_file)

        for f_img in doc.getElementsByTagName('Image'):
            f_image_id = \
                f_img.getElementsByTagName('ImageID')[0].childNodes[0].data
            if f_image_id == image_id:
                f_image = f_img
                break
            # end of if
        # end of for images

        if f_image is not None:
            for f_vol in f_image.getElementsByTagName('Volume'):
                f_volid = f_vol.getElementsByTagName('VolumeID')[0]
                f_volume_id = f_volid.childNodes[0].data
                if f_volume_id == volume_id:
                    f_volume = f_vol
                    break
                # end of if
            # end of for volumes
        # end of if

        if f_volume is not None:
            if remove is False:
                f_session = f_vol.getElementsByTagName('Session')

                if value is None:
                    f_session_num = \
                        str(int(f_session[0].childNodes[0].data) + 1)
                else:
                    f_session_num = value
                # end of if

                f_session[0].childNodes[0].data = f_session_num
                LOG.debug('_update_image_volume_info, update,'
                          'image_id:%(image_id)s,'
                          'volume_id:%(volume_id)s,'
                          'session_num:%(session_num)s'
                          % {'image_id': f_image_id,
                             'volume_id': f_volume_id,
                             'session_num': f_session_num})
            else:
                f_vol.previousSibling.data = ''
                f_img.removeChild(f_vol)
                LOG.debug('_update_image_volume_info, remove,'
                          'image_id:%(image_id)s,'
                          'volume_id:%(volume_id)s,'
                          % {'image_id': f_image_id,
                             'volume_id': f_volume_id})
            # end of if

            f = codecs.open(image_management_file, 'wb', 'UTF-8')
            doc.writexml(f, encoding='UTF-8')
            f.close()
        # end of if

        LOG.debug('_update_image_volume_info,Exit method')
        return

    def _get_sessionnum_by_srcvol(self, volume):
        '''
        get the number of session where specified volume is source
        '''
        LOG.debug('_get_sessionnum_by_srcvol,Enter method')

        session_info = []

        vol_instance = self._find_lun(volume)
        ResultClassStr = 'FUJITSU_StorageSynchronized'
        all_session_info = \
            self._reference_eternus_names(vol_instance.path,
                                          ResultClass=ResultClassStr)

        for session in all_session_info:
            if vol_instance.path != session['SyncedElement']:
                session_info.append(session)
            # end of if
        # end of for all_session_info

        session_num = len(session_info)
        LOG.debug('_get_sessionnum_by_srcvol,'
                  ' session_num:%(session_num)s,'
                  ' session_info:%(session_info)s'
                  % {'session_num': session_num,
                     'session_info': session_info})

        LOG.debug('_get_sessionnum_by_srcvol,Exit method')
        return session_num

    def _set_qos(self, volume):
        '''
        set qos using ETERNUS CLI
        '''
        LOG.debug('_set_qos,Enter method')

        category = 0

        qos_specs_dict = self._get_qos_specs(volume)
        for key, value in qos_specs_dict.iteritems():
            if key in FJ_QOS_KEY_list:
                category = self._get_qos_category_by_value(key, value)
                break
            # end of if
        # end of for qos_specs_dict

        if 0 < category:
            volumename = self._create_volume_name(volume['id'])
            param_dict = {'volume-name': volumename,
                          'bandwidth-limit': str(category)}
            rc, errordesc, job = self._exec_eternus_cli(
                'set_volume_qos',
                **param_dict)
            if rc != 0L:
                msg = (_LE('_set_qos,'
                           'Return code:%(rc)lu, '
                           'Error:%(errordesc)s, '
                           'Job:%(job)s')
                       % {'rc': rc,
                          'errordesc': errordesc,
                          'job': job})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if
        # end of if

        LOG.debug('_set_qos,Exit method')
        return

    def _get_qos_category_by_value(self, key, value):
        '''
        get qos category  using value
        '''
        LOG.debug('_get_qos_category_by_value,Enter method')

        ret = 0
        digit = 0

        # log error method
        def _get_qos_category_by_value_error():
            '''
            Input value is invalid,log error and raise exception
            '''
            msg = (_LE('_get_qos_category_by_value,'
                       'Invalid value is input,'
                       'key: %(key)s,'
                       'value: %(value)s')
                   % {'key': key,
                      'value': value})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of def

        if (key == "maxBWS"):
            try:
                digit = int(float(value))
            except:
                _get_qos_category_by_value_error()

            if 800 <= digit:
                ret = 1
            elif 700 <= digit:
                ret = 2
            elif 600 <= digit:
                ret = 3
            elif 500 <= digit:
                ret = 4
            elif 400 <= digit:
                ret = 5
            elif 300 <= digit:
                ret = 6
            elif 200 <= digit:
                ret = 7
            elif 100 <= digit:
                ret = 8
            elif 70 <= digit:
                ret = 9
            elif 40 <= digit:
                ret = 10
            elif 25 <= digit:
                ret = 11
            elif 20 <= digit:
                ret = 12
            elif 15 <= digit:
                ret = 13
            elif 10 <= digit:
                ret = 14
            elif 0 < digit:
                ret = 15
            else:
                _get_qos_category_by_value_error()
            # end of if
        # end of if

        LOG.debug('_get_qos_category_by_value (%s),Exit method' % str(ret))
        return ret

    def _get_target_port_info(self):
        '''
        return target_portid
        '''
        LOG.debug('_get_target_port_info,Enter method')

        lunportlist = []
        target_portlist = []
        target_cliportidlist = []

        lunportinslist = \
            self._enum_eternus_instances('FUJITSU_LUNMappingController')

        for lunportins in lunportinslist:
            lunportlist.append(
                self._conv_port_id(
                    lunportins['CeID'] if 'CeID' in lunportins else None,
                    lunportins['CMSlotNumber'],
                    lunportins['CASlotNumber'],
                    lunportins['PortNo']))
        # end of for lunportinslist

        if self.protocol == 'fc':
            prtcl_endpoint = 'FUJITSU_SCSIProtocolEndpoint'
            connection_type = 2
        elif self.protocol == 'iSCSI':
            prtcl_endpoint = 'FUJITSU_iSCSIProtocolEndpoint'
            connection_type = 7

        try:
            tgtportlist = self._enum_eternus_instances(prtcl_endpoint)
        except:
            msg = (_LE('_get_target_port_info,'
                       'EnumerateInstances,'
                       'cannot connect to ETERNUS.'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for tgtport in tgtportlist:
            cliportid = \
                self._conv_port_id(
                    tgtport['CeID'] if 'CeID' in tgtport else None,
                    tgtport['CMSlotNumber'],
                    tgtport['CASlotNumber'],
                    tgtport['PortNo'])
            if tgtport['ConnectionType'] != connection_type:
                continue
            # end of if
            if (tgtport['RAMode'] & 0x7B) != 0x00:
                continue
            # end of if
            if 'SCGroupNo' in tgtport:
                continue
            # end of if
            if cliportid in lunportlist:
                continue
            # end of if
            target_portlist.append(tgtport)
            target_cliportidlist.append(cliportid)

            LOG.debug('_get_target_port_info,'
                      'connection type:%(cont)s,'
                      'ramode:%(ramode)s'
                      % {'cont': tgtport['ConnectionType'],
                         'ramode': tgtport['RAMode']})
        # end of for tgtportlist

        LOG.debug('_get_target_port_info,'
                  'target port: %(target_portid)s '
                  % {'target_portid': target_portlist})
        LOG.debug('_get_target_port_info,'
                  'target portid: %(target_portid)s '
                  % {'target_portid': target_cliportidlist})
        # end of if

        if len(target_portlist) == 0:
            msg = (_LE('_get_target_port_info,'
                       'protcol:%(protocol)s,'
                       'target_port does not found.')
                   % {'protocol': self.protocol})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('_get_target_port_info,Exit method')
        return (target_portlist, target_cliportidlist)

    @eternus_dx_lockutils.FJDXLockutils('connect', 'cinder-', True)
    def _map_lun(self, vol_instance, connector, portidlist=[]):
        '''
        map volume to host.
        '''

        volumename = vol_instance['ElementName']
        LOG.debug('_map_lun,'
                  'volumename:%(volumename)s,'
                  'connector:%(con)s,'
                  'Enter method'
                  % {'volumename': volumename,
                     'con': connector})

        devid_preset = set()

        volume_uid = vol_instance['Name']
        volume_lun = vol_instance['LUN']
        initiatorlist = self._find_initiator_names(connector)
        aglist = self._find_affinity_group(connector)
        configservice = self._find_eternus_service(CTRL_CONF)

        if len(portidlist) == 0:
            _x, portidlist = self._get_target_port_info()
        # end of if

        if configservice is None:
            msg = (_LE('_map_lun,'
                       'vol_instance.path:%(vol)s,'
                       'volumename:%(volumename)s,'
                       'volume_uid:%(uid)s,'
                       'volume_lun:%(lun)s,'
                       'initiator:%(initiator)s,'
                       'portid:%(portid)s,'
                       'aglist:%(aglist)s,'
                       'Cannot find Controller Configuration')
                   % {'vol': str(vol_instance.path),
                      'volumename': [volumename],
                      'uid': [volume_uid],
                      'lun': [volume_lun],
                      'initiator': initiatorlist,
                      'portid': portidlist,
                      'aglist': aglist})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if
        LOG.debug('_map_lun,'
                  'vol_instance.path:%(vol)s,'
                  'volumename:%(volumename)s,'
                  'initiator:%(initiator)s,'
                  'portid:%(portid)s'
                  % {'vol': str(vol_instance.path),
                     'volumename': [volumename],
                     'initiator': initiatorlist,
                     'portid': portidlist})

        if len(aglist) == 0:
            initiatorlistwk = [sss.lower() for sss in initiatorlist]
            hostnolist = []
            command = None
            hostname = None
            agnum = None

            if 'SCGroupNo' in vol_instance:
                msg = (_LE('_map_lun,'
                           'vol_instance.path:%(vol)s,'
                           'volumename:%(volumename)s,'
                           'volume_uid:%(uid)s,'
                           'volume_lun:%(lun)s,'
                           '%(errordesc)s')
                       % {'vol': str(vol_instance.path),
                          'volumename': [volumename],
                          'uid': [volume_uid],
                          'lun': [volume_lun],
                          'errordesc': RETCODE_dic['32812']})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            ########################################
            # get host name
            command = None

            if self.protocol == 'fc':
                command = 'show_host_wwn_names'
            elif self.protocol == 'iSCSI':
                command = 'show_host_iscsi_names'

            # end of if

            if command:
                rc, emsg, clidata = self._exec_eternus_cli(command)

                if rc != 0L:
                    msg = (_LE('_map_lun,'
                               'Return code:%(rc)lu, '
                               'Error code:%(clidata)s, '
                               'Message:%(emsg)s')
                           % {'rc': rc,
                              'clidata': clidata,
                              'emsg': emsg})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                # end of if

                for hostdata in clidata:
                    if hostdata['Host Name'].lower() in initiatorlistwk:
                        hostnolist.append(str(hostdata['Host Num']))
                        initiatorlistwk.remove(hostdata['Host Name'].lower())
                    # end of if
                # end of for clidata
            # end of if

            ########################################
            # create new host
            if self.protocol == 'fc':
                command = 'create_host_wwn_name'
                hostname = 'wwn'
            elif self.protocol == 'iSCSI':
                command = 'create_host_iscsi_name'
                hostname = 'iscsi-name'
            # end of if

            if command:
                for initiator in initiatorlist:
                    if initiator.lower() in initiatorlistwk:
                        option = {hostname: initiator}
                        rc, emsg, clidata = self._exec_eternus_cli(command,
                                                                   **option)

                        if rc != 0L:
                            msg = (_LE('_map_lun,'
                                       'Return code:%(rc)lu, '
                                       'Error code:%(clidata)s, '
                                       'Message:%(emsg)s')
                                   % {'rc': rc,
                                      'clidata': clidata,
                                      'emsg': emsg})
                            LOG.error(msg)
                            raise exception.VolumeBackendAPIException(data=msg)
                        try:
                            hostnolist.append(str(int(clidata[0], 16)))
                        except:
                            msg = (_LE('_map_lun,'
                                       'Return code:%(rc)lu, '
                                       'Error code:%(clidata)s, '
                                       'Message:%(emsg)s')
                                   % {'rc': rc,
                                      'clidata': clidata,
                                      'emsg': emsg})
                            LOG.error(msg)
                            raise exception.VolumeBackendAPIException(data=msg)
                        # end of if
                    # end of if
                # end of for initiatorlist
            # end of if

            ########################################
            # create new affinity group
            option = {'volume-number': volume_lun,
                      'lun': '0'}
            rc, emsg, clidata = self._exec_eternus_cli('create_affinity_group',
                                                       **option)

            if rc != 0L:
                msg = (_LE('_map_lun,'
                           'Return code:%(rc)lu, '
                           'Error code:%(clidata)s, '
                           'Message:%(emsg)s')
                       % {'rc': rc,
                          'clidata': clidata,
                          'emsg': emsg})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            try:
                agnum = int(clidata[0], 16)
            except:
                msg = (_LE('_map_lun,'
                           'Return code:%(rc)lu, '
                           'Error code:%(clidata)s, '
                           'Message:%(emsg)s')
                       % {'rc': rc,
                          'clidata': clidata,
                          'emsg': emsg})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if hostnolist:
                ########################################
                # set host affinity
                option = {'ag-number': agnum,
                          'host-number': ','.join(hostnolist),
                          'port': ','.join(portidlist)}
                rc, emsg, clidata = self._exec_eternus_cli('set_host_affinity',
                                                           **option)

                if rc != 0L:
                    msg = (_LE('_map_lun,'
                               'Return code:%(rc)lu, '
                               'Error code:%(clidata)s, '
                               'Message:%(emsg)s')
                           % {'rc': rc,
                              'clidata': clidata,
                              'emsg': emsg})
                    LOG.error(msg)

                    ########################################
                    # delete affinity group
                    option = {'ag-number': agnum}
                    rc, emsg, clidata = \
                        self._exec_eternus_cli('delete_affinity_group',
                                               **option)

                    if rc != 0L:
                        msg = (_LE('_map_lun,'
                                   'Return code:%(rc)lu, '
                                   'Error code:%(clidata)s, '
                                   'Message:%(emsg)s')
                               % {'rc': rc,
                                  'clidata': clidata,
                                  'emsg': emsg})
                        LOG.error(msg)
                    # end of if

                    raise exception.VolumeBackendAPIException(data=msg)
                # end of if
            # end of if
        else:
            # add lun to affinity group
            for ag in aglist:
                LOG.debug('_map_lun,'
                          'ag:%(ag)s,'
                          'lun_name:%(volume_uid)s'
                          % {'ag': str(ag),
                             'volume_uid': volume_uid})

                devid_pre = ag['DeviceID'][:8]

                if devid_pre not in devid_preset:
                    rc, errordesc, job = self._exec_eternus_service(
                        'ExposePaths',
                        configservice, LUNames=[volume_uid],
                        DeviceAccesses=[pywbem.Uint16(2)],
                        ProtocolControllers=[ag])

                    LOG.debug('_map_lun,'
                              'Error:%(errordesc)s,'
                              'Return code:%(rc)lu,'
                              'Add lun affinitygroup'
                              % {'errordesc': errordesc,
                                 'rc': rc})

                    if rc != 0L:
                        msg = (_LE('_map_lun,'
                                   'lun_name:%(volume_uid)s,'
                                   'Initiator:%(initiator)s,'
                                   'portid:%(portid)s,'
                                   'Return code:%(rc)lu,'
                                   'Error:%(errordesc)s')
                               % {'volume_uid': [volume_uid],
                                  'initiator': initiatorlist,
                                  'portid': portidlist,
                                  'rc': rc,
                                  'errordesc': errordesc})
                        LOG.error(msg)
                    # end of if

                    devid_preset.add(devid_pre)
                # end of if
            # end of for aglist
        # end of if
        LOG.debug('_map_lun,'
                  'volumename:%(volumename)s,'
                  'Exit method'
                  % {'volumename': volumename})
        return

    def _check_user(self):
        '''
        check whether user's role is accessible to ETERNUS and Software
        '''
        LOG.debug('_check_user,Enter method')

        ret = True
        rc, errordesc, job = self._exec_eternus_cli('check_user_role')
        if rc != 0L:
            msg = (_LE('_check_user,'
                       'Return code:%(rc)lu, '
                       'Error:%(errordesc)s, '
                       'Job:%(job)s')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'job': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        if job != 'Software':
            msg = (_LE('_check_user,'
                       'Specified user(%(user)s) does not have '
                       'Software role: %(role)s')
                   % {'user': self._get_drvcfg('EternusUser'),
                      'role': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        # end of if

        LOG.debug('_check_user,Exit method')
        return ret

    @eternus_dx_lockutils.FJDXLockutils('SMIS-exec', 'cinder-', True)
    def _exec_eternus_cli(self, command, retry=20, retry_interval=5,
                          retry_code=[32787L], **param_dict):
        '''
        Execute ETERNUS CLI
        '''
        LOG.debug('_exec_eternus_cli,'
                  'command:%(a)s,'
                  'parameters:%(b)s,'
                  'Enter method'
                  % {'a': str(command),
                     'b': str(param_dict)})

        out = None
        _err = None
        result = None
        rc = None
        retdata = None
        errordesc = None
        user = self._get_drvcfg('EternusUser')
        storage_ip = self._get_drvcfg('EternusIP')

        for retry_num in range(retry):
            # execute ETERNUS CLI & get return value
            try:
                if param_dict:
                    (out, _err) = utils.execute('eternus_dx_cli', '-u', user,
                                                '-l', storage_ip,
                                                '-c', command,
                                                '-o', str(param_dict),
                                                run_as_root=True)
                else:
                    (out, _err) = utils.execute('eternus_dx_cli', '-u', user,
                                                '-l', storage_ip,
                                                '-c', command,
                                                run_as_root=True)
                # end of if

                out_dict = eval(out)
                result = out_dict.get('result')
                rc_str = out_dict.get('rc')
                retdata = out_dict.get('message')
            except Exception as ex:
                msg = (_LE('_exec_eternus_cli,'
                           'stdout: %(out)s,'
                           'stderr: %(err)s,'
                           'unexpected error: %(ex)s,')
                       % {'out': out,
                          'err': _err,
                          'ex': str(ex)})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # check ssh result
            if result == 255:
                LOG.info(_LI('_exec_eternus_cli, retry,'
                             'command:%(command)s,'
                             'option:%(option)s,'
                             'ip:%(ip)s,'
                             'SSH Result:%(result)s,'
                             'retdata:%(retdata)s,'
                             'TryNum:%(rn)s')
                         % {'command': command,
                            'option': str(param_dict),
                            'ip': storage_ip,
                            'result': str(result),
                            'retdata': retdata,
                            'rn': str(retry_num + 1)})
                time.sleep(retry_interval)
                continue
            elif result != 0:
                msg = (_LE('_exec_eternus_cli,'
                           'unexpected error,'
                           'command:%(command)s,'
                           'option:%(option)s,'
                           'ip:%(ip)s,'
                           'result:%(result)s,'
                           'retdata:%(retdata)s')
                       % {'command': command,
                          'option': str(param_dict),
                          'ip': storage_ip,
                          'result': str(result),
                          'retdata': retdata})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            # check CLI return code
            if rc_str.isdigit():
                # SMI-S style return code
                rc = int(rc_str)

                try:
                    errordesc = RETCODE_dic[str(rc)]
                except:
                    errordesc = 'Undefined Error!!'

                if rc in retry_code:
                    LOG.info(_LI('_exec_eternus_cli, retry,'
                                 'ip:%(ip)s,'
                                 'RetryCode:%(rc)s,'
                                 'TryNum:%(rn)s')
                             % {'ip': storage_ip,
                                'rc': str(rc),
                                'rn': str(retry_num + 1)})
                    time.sleep(retry_interval)
                    continue
                # end of if

                break
            else:
                # CLI style return code
                LOG.warn(_LW('_exec_eternus_cli,'
                             'WARNING!!,'
                             'ip:%(ip)s,'
                             'ReturnCode:%(rc_str)s,'
                             'ReturnData:%(retdata)s')
                         % {'ip': storage_ip,
                            'rc_str': rc_str,
                            'retdata': retdata})

                errordesc = retdata
                retdata = rc_str
                rc = 4  # Failed
                break
            # end of if
        else:
            if 0 < result:
                msg = (_LE('_exec_eternus_cli,'
                           'cannot connect to ETERNUS.'
                           'SSH Result:%(result)s,'
                           'retdata:%(retdata)s')
                       % {'result': str(result),
                          'retdata': retdata,
                          'rn': str(retry_num + 1)})

                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                LOG.warn(_LW('_exec_eternus_cli, Retry was exceeded.'))
            # end of if
        # end of for retry

        ret = (rc, errordesc, retdata)

        LOG.debug('_exec_eternus_cli,'
                  'command:%(a)s,'
                  'parameters:%(b)s,'
                  'ip:%(ip)s,'
                  'Return code:%(rc)s,'
                  'Error:%(errordesc)s,'
                  'Exit method'
                  % {'a': str(command),
                     'b': str(param_dict),
                     'ip': storage_ip,
                     'rc': rc,
                     'errordesc': errordesc})
        return ret
