# Copyright (c) 2015 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

'''
iSCSI Cinder Volume driver for Fujitsu ETERNUS DX S3 series.
This driver is extension version.
'''

# $Date:: 2015-09-17 18:59:39 +0900#$
# $Revision: 194 $

from oslo_config import cfg
from cinder.volume.drivers.fujitsu.extension import image_utils
from oslo_log import log as logging
from cinder import exception
from cinder import context
from cinder import utils
from cinder.volume.drivers.fujitsu import eternus_dx_iscsi
from cinder.volume.drivers.fujitsu.extension import eternus_dx_common
from oslo_concurrency import lockutils
import threading
import six
from cinder.i18n import _LE, _LI, _LW

LOG = logging.getLogger(__name__)

FJ_SRC_VOL_ID = "FJ_Source_Volume_ID"
FJ_REMOTE_DEST_META = "FJ_Remote_Copy_Destination"
FJ_ETERNUS_DX_OPT_list = [cfg.BoolOpt('use_fujitsu_image_volume',
                                      default=True,
                                      help='Use image volume when '
                                      'copying image to volume')]


class FJDXISCSIDriver(eternus_dx_iscsi.FJDXISCSIDriver):
    '''
    ETERNUS iSCSI Cinder Volume Driver
    for Fujitsu ETERNUS DX S3 series
    '''

    def __init__(self, *args, **kwargs):

        super(FJDXISCSIDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(FJ_ETERNUS_DX_OPT_list)
        self.common = \
            eternus_dx_common.FJDXCommonExt(
                'iSCSI', configuration=self.configuration)
        self.first_loop = True
        self.VERSION = self.common.VERSION

        return

    def create_volume(self, volume):
        '''
        Create volume
        '''
        LOG.info(_LI('create_volume,Enter method'))
        element_path = None
        metadata = None
        d_metadata = {}

        # get volume metadata
        v_metadata = volume.get('volume_metadata')
        for data in v_metadata:
            d_metadata[data['key']] = data['value']
        # end of for v_metadata

        # choice : standard create volume or CCM create volume
        if FJ_SRC_VOL_ID not in d_metadata:
            # standard create volume
            (element_path, metadata) = \
                self.common.create_volume(volume)
        else:
            # CCM create volume
            # create cloned volume for backup solution
            # get source & target volume information
            ctxt = context.get_admin_context()
            volume_size = volume['size']
            src_volid = d_metadata[FJ_SRC_VOL_ID]
            if src_volid is None:
                msg = (_LE('create_volume,Source volume id is None'))
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            # modify metadata value Mirror to 'mirror'
            if FJ_REMOTE_DEST_META in d_metadata:
                d_metadata[FJ_REMOTE_DEST_META] = \
                    (d_metadata[FJ_REMOTE_DEST_META]).lower()

            # get info of source volume
            src_vref = self.db.volume_get(ctxt, src_volid)

            # check target volume information with reference to source volume
            src_vref_size = src_vref['size']
            src_vref_status = src_vref['status']

            if ('error' in src_vref_status) or ('deleting' in src_vref_status):
                msg = (_LE('Invalid volume status : %(status)s')
                       % {'status': src_vref_status})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            elif volume_size < src_vref_size:
                msg = (_LE('Volume size %(volume_size)sGB '
                           'cannot be smaller than '
                           'original volume size %(src_vref_size)sGB.'
                           'They must be >= original volume size.')
                       % {'volume_size': volume_size,
                          'src_vref_size': src_vref_size})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            # end of if

            LOG.info(_LI('create_volume,Call create cloned volume'))
            (element_path, metadata) = \
                self.common.create_cloned_volume(volume,
                                                 src_vref,
                                                 use_service_name=True)
            LOG.info(_LI('create_volume,Return create cloned volume'))

            # post processing
            try:
                if src_vref.bootable:
                    self.db.volume_update(ctxt,
                                          volume['id'],
                                          {'bootable': True})
                    self.db.volume_glance_metadata_copy_from_volume_to_volume(
                        ctxt, src_volid, volume['id'])
                # end of if

                self.db.volume_update(ctxt,
                                      volume['id'],
                                      {'source_volid': src_volid})
            except Exception as ex:
                msg = ('create volume, Failed updating volume metadata,'
                       'reason: %(reason)s,volume_id: %(volume_id)s,'
                       'src_volid: %(src_volid)s,'
                       % {'reason': ex,
                          'volume_id': volume['id'],
                          'src_volid': src_volid})
                raise exception.MetadataCopyFailure(reason=msg)
        # end of if

        metadata.update(d_metadata)

        LOG.info(_LI('create_volume,Exit method'))
        return {'provider_location': six.text_type(element_path),
                'metadata': metadata}

    def create_volume_from_snapshot(self, volume, snapshot):
        '''
        Creates a volume from a snapshot.
        '''
        LOG.info(_LI('create_volume_from_snapshot,Enter method'))

        (element_path, metadata) = \
            self.common.create_volume_from_snapshot(volume, snapshot)

        v_metadata = volume.get('volume_metadata')
        for data in v_metadata:
            metadata[data['key']] = data['value']
        # end of for v_metadata

        LOG.info(_LI('create_volume_from_snapshot,Exit method'))
        return {'provider_location': six.text_type(element_path),
                'metadata': metadata}

    def create_cloned_volume(self, volume, src_vref):
        '''
        Create cloned volume.
        '''
        LOG.info(_LI('create_cloned_volume,Enter method'))

        (element_path, metadata) = \
            self.common.create_cloned_volume(volume, src_vref)

        v_metadata = volume.get('volume_metadata')
        for data in v_metadata:
            metadata[data['key']] = data['value']
        # end of for v_metadata

        LOG.info(_LI('create_cloned_volume,Exit method'))
        return {'provider_location': six.text_type(element_path),
                'metadata': metadata}

    def delete_volume(self, volume):
        '''
        Delete volume on ETERNUS.
        '''
        LOG.info(_LI('delete_volume,Enter method'))

        self.common.delete_volume(volume)

        LOG.info(_LI('delete_volume,Exit method'))
        return

    def initialize_connection(self, volume, connector):
        '''
        Allow connection to connector and return connection info.
        '''
        LOG.info(_LI('initialize_connection,'
                     'volume id: %(volid)s,'
                     'connector info: %(initiator)s,Enter method')
                 % {'volid': volume['id'],
                    'initiator': connector['initiator']})

        info = self.common.initialize_connection(volume, connector)

        LOG.info(_LI('initialize_connection,'
                     'connection info: %(info)s'
                     'Exit method') % {'info': info})
        return info

    def terminate_connection(self, volume, connector, **kwargs):
        '''
        Disallow connection from connector.
        '''
        LOG.info(_LI('terminate_connection,'
                     'volume id: %(volid)s,'
                     'connector info: %(initiator)s,Enter method')
                 % {'volid': volume['id'],
                    'initiator': connector['initiator']})

        self.common.terminate_connection(volume, connector)

        LOG.info(_LI('terminate_connection,Exit method'))
        return

    def get_volume_stats(self, refresh=False):
        '''
        Get volume stats.
        If 'refresh' is True, run update the stats first.
        '''
        LOG.debug('get_volume_stats,Enter method')

        if refresh is True:
            data = self.common.update_volume_stats()
            backend_name = self.configuration.safe_get('volume_backend_name')
            data['volume_backend_name'] = backend_name or 'FJDXISCSIDriver'
            data['storage_protocol'] = 'iSCSI'
            self._stats = data
        # end of if

        LOG.debug('get_volume_stats,Exit method')

        if self.first_loop is True:
            self.first_loop = False
            monitor_thread = \
                threading.Thread(target=self.common.monitor_image_volume)
            monitor_thread.start()
        # end of if

        return self._stats

    def extend_volume(self, volume, new_size):
        '''
        Extend volume.
        '''
        LOG.info(_LI('extend_volume,Enter method'))

        self.common.extend_volume(volume, new_size)

        LOG.info(_LI('extend_volume,Exit method'))
        return

    @lockutils.synchronized('ETERNUS_DX-img', 'cinder-', True)
    def copy_image_to_volume(self, context, volume, image_service, image_id):
        '''
        Fetch the image from image_service and write it to the volume.
        '''
        LOG.info(_LI('copy_image_to_volume,'
                     'Enter method (use_image_volume: %s)')
                 % self.configuration.use_fujitsu_image_volume)

        cloned = False
        image_volume = None

        setup = self._setup_image_volume(volume)

        use_multipath = self.configuration.use_multipath_for_image_xfer
        enforce_multipath = self.configuration.enforce_multipath_for_image_xfer
        properties = utils.brick_get_connector_properties(
            use_multipath, enforce_multipath)

        if self.configuration.use_fujitsu_image_volume is True:

            try:  # creating image volume using dd command
                if setup is False:
                    cloned = self._copy_image_volume_to_volume(volume,
                                                               image_id)
                # end of if

                if cloned is True:
                    LOG.info(_LI('copy_image_to_volume,'
                                 'create volume using OPC'))
                else:
                    image_volume = self._create_image_volume(volume, image_id)

                    try:
                        self._write_image_to_volume(context,
                                                    properties,
                                                    image_volume,
                                                    image_service,
                                                    image_id)
                        self._add_image_volume(image_volume, image_id)
                    except Exception as ex:
                        LOG.info(_LI(ex))
                        self.delete_volume(image_volume)
                        raise ex

                    if setup is False:
                        cloned = self._copy_image_volume_to_volume(
                            volume, image_id, src_vref=image_volume)
                        LOG.info(_LI('copy_image_to_volume, create '
                                     'volume by writing image, '
                                     'result : %s') % (cloned))
                    # end of if
                # end of if
            except Exception as ex:
                LOG.warn(_LW(ex))
                LOG.warn(_LW('copy_image_to_volume,create volume '
                             'without creating image volume'))
                if setup is False:
                    self._write_image_to_volume(context, properties,
                                                volume, image_service,
                                                image_id)
                else:
                    raise ex
                # end of if
        elif setup is True:  # use_fujitsu_image_volume=False & setup=True
            msg = (_LE('copy_image_to_volume, this backend specify '
                       'that use_fujitsu_image_volume is False'))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        else:  # if not self.configuration.use_fujitsu_image_volume is True
            properties = utils.brick_get_connector_properties(
                use_multipath, enforce_multipath)
            self._write_image_to_volume(context, properties,
                                        volume, image_service, image_id)
        # end of if

        LOG.info(_LI('copy_image_to_volume,Exit method'))
        return

    def _write_image_to_volume(self, context, properties,
                               volume, image_service, image_id):
        '''
        Fetch the image from image_service and write it to the volume
        using dd command.
        '''
        LOG.debug('copy_image_to_volume,Enter method')

        attach_info, volume = self._attach_volume(context, volume, properties)

        try:
            image_utils.fetch_to_raw(context,
                                     image_service,
                                     image_id,
                                     attach_info['device']['path'],
                                     self.configuration.volume_dd_blocksize,
                                     size=volume['size'])
        finally:
            self._detach_volume(context, attach_info, volume, properties)

        LOG.debug('copy_image_to_volume,Exit method')
        return

    def _setup_image_volume(self, volume):
        '''
        Check whether setup mode or not, and prepare for setup mode
        '''
        LOG.debug('_setup_image_volume,Enter method')

        setup = False

        meta = volume.get('volume_metadata')
        for data in meta:
            if data['key'] == 'fujitsu_image_setup':
                if data['value'] == 'True':
                    setup = True
                    volume_type_id = volume['volume_type_id']
                    volume['volume_type_id'] = None
                    self.common.delete_volume(volume)
                    volume['volume_type_id'] = volume_type_id
                    LOG.info(_LI('_setup_image_volume, setup:%s') % (setup))
                    break
                # end of if
            # end of if
        # end of for meta

        LOG.debug('_setup_image_volume (setup:%s),Exit method' % (setup))
        return setup

    def _create_image_volume(self, volume_ref, image_id):
        '''
        Create blank volume for image volume
        '''
        LOG.debug('_create_image_volume,Enter method')

        image_volume = self.common.create_image_volume(volume_ref, image_id)

        LOG.debug('_create_image_volume,Exit method')
        return image_volume

    def _add_image_volume(self, volume_ref, image_id):
        '''
        Create blank volume for image volume
        '''
        LOG.debug('_add_image_volume,Enter method')

        self.common.add_image_volume(volume_ref, image_id)

        LOG.debug('_add_image_volume,Exit method')
        return

    def _copy_image_volume_to_volume(self, volume, image_id, src_vref=None):
        '''
        Copy image volume to volume
        '''
        LOG.debug('_copy_image_volume_to_volume,Enter method')

        cloned = self.common.copy_image_volume_to_volume(volume,
                                                         image_id,
                                                         src_vref)

        LOG.debug('_copy_image_volume_to_volume (cloned:%s),'
                  'Exit method' % (cloned))
        return cloned
